"""
on_ticket is the main function that is called when a new issue is created.
It is only called by the webhook handler in sweepai/api.py.
"""

import difflib
import io
import os
import re
import traceback
from typing import Any
import zipfile
from time import time

import markdown
import openai
import requests
import yaml
import yamllint.config as yamllint_config
from github import BadCredentialsException, Github, Repository
from github.Issue import Issue
from github.PullRequest import PullRequest as GithubPullRequest
from loguru import logger
from tabulate import tabulate
from tqdm import tqdm
from yamllint import linter

from sweepai.agents.pr_description_bot import PRDescriptionBot
from sweepai.config.client import (
    DEFAULT_RULES,
    RESET_FILE,
    RESTART_SWEEP_BUTTON,
    REVERT_CHANGED_FILES_TITLE,
    RULES_LABEL,
    RULES_TITLE,
    SWEEP_BAD_FEEDBACK,
    SWEEP_GOOD_FEEDBACK,
    SweepConfig,
    get_documentation_dict,
    get_gha_enabled,
    get_rules,
)
from sweepai.config.server import (
    DEPLOYMENT_GHA_ENABLED,
    DISCORD_FEEDBACK_WEBHOOK_URL,
    ENV,
    GITHUB_LABEL_NAME,
    IS_SELF_HOSTED,
    MONGODB_URI,
    PROGRESS_BASE_URL,
)
from sweepai.core.context_pruning import get_relevant_context
from sweepai.core.entities import (
    AssistantRaisedException,
    FileChangeRequest,
    MaxTokensExceeded,
    NoFilesException,
    PullRequest,
    SandboxResponse,
)
from sweepai.core.entities import create_error_logs as entities_create_error_logs
from sweepai.core.pr_reader import PRReader
from sweepai.core.sweep_bot import SweepBot, get_files_to_change, validate_file_change_requests
from sweepai.handlers.create_pr import (
    create_config_pr,
    create_pr_changes,
    safe_delete_sweep_branch,
)
from sweepai.handlers.on_check_suite import clean_gh_logs
from sweepai.utils.buttons import Button, ButtonList, create_action_buttons
from sweepai.utils.chat_logger import ChatLogger
from sweepai.utils.diff import generate_diff
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import (
    CURRENT_USERNAME,
    ClonedRepo,
    convert_pr_draft_field,
    get_github_client,
    get_token,
)
from sweepai.utils.progress import (
    AssistantConversation,
    PaymentContext,
    TicketContext,
    TicketProgress,
    TicketProgressStatus,
)
from sweepai.utils.prompt_constructor import HumanMessagePrompt
from sweepai.utils.str_utils import (
    BOT_SUFFIX,
    FASTER_MODEL_MESSAGE,
    UPDATES_MESSAGE,
    blockquote,
    bot_suffix,
    checkbox_template,
    clean_logs,
    collapsible_template,
    create_checkbox,
    create_collapsible,
    discord_suffix,
    format_sandbox_success,
    get_hash,
    sep,
    stars_suffix,
    strip_sweep,
    to_branch_name,
)
from sweepai.utils.ticket_utils import (
    center,
    fetch_relevant_files,
    fire_and_forget_wrapper,
    log_error,
    prep_snippets,
)
from sweepai.utils.user_settings import UserSettings

# from sandbox.sandbox_utils import Sandbox


sweeping_gif = """<a href="https://github.com/sweepai/sweep"><img class="swing" src="https://raw.githubusercontent.com/sweepai/sweep/main/.assets/sweeping.gif" width="100" style="width:50px; margin-bottom:10px" alt="Sweeping"></a>"""


custom_config = """
extends: relaxed

rules:
    line-length: disable
    indentation: disable
"""

INSTRUCTIONS_FOR_REVIEW = """\
### 💡 To get Sweep to edit this pull request, you can:
* Comment below, and Sweep can edit the entire PR
* Comment on a file, Sweep will only modify the commented file
* Edit the original issue to get Sweep to recreate the PR from scratch"""

email_template = """Hey {name},
<br/><br/>
🚀 I just finished creating a pull request for your issue ({repo_full_name}#{issue_number}) at <a href="{pr_url}">{repo_full_name}#{pr_number}</a>!

<br/><br/>
You can view how I created this pull request <a href="{progress_url}">here</a>.

<h2>Summary</h2>
<blockquote>
{summary}
</blockquote>

<h2>Files Changed</h2>
<ul>
{files_changed}
</ul>

{sweeping_gif}
<br/>
Cheers,
<br/>
Sweep
<br/>"""

FAILING_GITHUB_ACTION_PROMPT = """\
The following Github Actions failed on a previous attempt at fixing this issue.
Propose a fix to the failing github actions. You must edit the source code, not the github action itself.
{github_action_log}
"""


# Add :eyes: emoji to ticket
def add_emoji(issue: Issue, comment_id: int = None, reaction_content="eyes"):
    item_to_react_to = issue.get_comment(comment_id) if comment_id else issue
    item_to_react_to.create_reaction("eyes")


# If SWEEP_BOT reacted to item_to_react_to with "rocket", then remove it.
def remove_emoji(issue: Issue, comment_id: int = None, content_to_delete="eyes"):
    item_to_react_to = issue.get_comment(comment_id) if comment_id else issue
    reactions = item_to_react_to.get_reactions()
    for reaction in reactions:
        if (
            reaction.content == content_to_delete
            and reaction.user.login == CURRENT_USERNAME
        ):
            item_to_react_to.delete_reaction(reaction.id)


def create_error_logs(
    commit_url_display: str,
    sandbox_response: SandboxResponse,
    status: str = "✓",
):
    return (
        (
            "<br/>"
            + create_collapsible(
                f"Sandbox logs for {commit_url_display} {status}",
                blockquote(
                    "\n\n".join(
                        [
                            create_collapsible(
                                f"<code>{output}</code> {i + 1}/{len(sandbox_response.outputs)} {format_sandbox_success(sandbox_response.success)}",
                                f"<pre>{clean_logs(output)}</pre>",
                                i == len(sandbox_response.outputs) - 1,
                            )
                            for i, output in enumerate(sandbox_response.outputs)
                            if len(sandbox_response.outputs) > 0
                        ]
                    )
                ),
                opened=True,
            )
        )
        if sandbox_response
        else ""
    )


# takes in a list of workflow runs and returns a list of messages containing the logs of the failing runs
def get_failing_gha_logs(runs, installation_id) -> str:
    token = get_token(installation_id)
    all_logs = ""
    for run in runs:
        # jobs_url
        jobs_url = run.jobs_url
        jobs_response = requests.get(
            jobs_url,
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": f"Bearer {token}",
                "X-GitHub-Api-Version": "2022-11-28",
            },
        )
        if jobs_response.status_code == 200:
            failed_jobs = []
            jobs = jobs_response.json()["jobs"]
            for job in jobs:
                if job["conclusion"] == "failure":
                    failed_jobs.append(job)

            failed_jobs_name_list = []
            for job in failed_jobs:
                # add failed steps
                for step in job["steps"]:
                    if step["conclusion"] == "failure":
                        failed_jobs_name_list.append(
                            f"{job['name']}/{step['number']}_{step['name']}"
                        )
        else:
            logger.error(
                "Failed to get jobs for failing github actions, possible a credentials issue"
            )
            return all_logs
        # logs url
        logs_url = run.logs_url
        logs_response = requests.get(
            logs_url,
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": f"Bearer {token}",
                "X-GitHub-Api-Version": "2022-11-28",
            },
            allow_redirects=True,
        )
        # Check if the request was successful
        if logs_response.status_code == 200:
            zip_data = io.BytesIO(logs_response.content)
            zip_file = zipfile.ZipFile(zip_data, "r")
            zip_file_names = zip_file.namelist()
            for file in failed_jobs_name_list:
                if f"{file}.txt" in zip_file_names:
                    logs = zip_file.read(f"{file}.txt").decode("utf-8")
                    logs_prompt = clean_gh_logs(logs)
                    all_logs += logs_prompt + "\n"
        else:
            logger.error(
                "Failed to get logs for failing github actions, likely a credentials issue"
            )
            return all_logs
    return all_logs


def delete_old_prs(repo: Repository, issue_number: int):
    logger.info("Deleting old PRs...")
    prs = repo.get_pulls(
        state="open",
        sort="created",
        direction="desc",
        base=SweepConfig.get_branch(repo),
    )
    for pr in tqdm(prs.get_page(0)):
        # # Check if this issue is mentioned in the PR, and pr is owned by bot
        # # This is done in create_pr, (pr_description = ...)
        if pr.user.login == CURRENT_USERNAME and f"Fixes #{issue_number}.\n" in pr.body:
            safe_delete_sweep_branch(pr, repo)
            break

def construct_sweep_bot(
        repo: Repository,
        repo_name: str,
        issue_url: str,
        repo_description: str,
        title: str,
        message_summary: str,
        cloned_repo: ClonedRepo,
        ticket_progress: TicketProgress,
        chat_logger: ChatLogger,
        snippets: Any = None,
        tree: Any = None,
        comments: Any = None,
    ) -> SweepBot:
    human_message = HumanMessagePrompt(
        repo_name=repo_name,
        issue_url=issue_url,
        repo_description=repo_description.strip(),
        title=title,
        summary=message_summary,
        snippets=snippets,
        tree=tree,
    )
    sweep_bot = SweepBot.from_system_message_content(
        human_message=human_message,
        repo=repo,
        is_reply=bool(comments),
        chat_logger=chat_logger,
        cloned_repo=cloned_repo,
        ticket_progress=ticket_progress,
    )
    return sweep_bot


def get_comment_header(
    index: int,
    g: Github,
    repo_full_name: str,
    user_settings: UserSettings,
    progress_headers: list[None | str],
    tracking_id: str | None,
    payment_message_start: str,
    user_settings_message: str,
    errored: bool = False,
    pr_message: str = "",
    done: bool = False,
    initial_sandbox_response: int | SandboxResponse = -1,
    initial_sandbox_response_file=None,
    config_pr_url: str | None = None,
):
    config_pr_message = (
        "\n"
        + f"<div align='center'>Install Sweep Configs: <a href='{config_pr_url}'>Pull Request</a></div>"
        if config_pr_url is not None
        else ""
    )
    actions_message = create_action_buttons(
        [
            RESTART_SWEEP_BUTTON,
        ]
    )

    sandbox_execution_message = "\n\n## GitHub Actions failed\n\nThe sandbox appears to be unavailable or down.\n\n"

    if initial_sandbox_response == -1:
        sandbox_execution_message = ""
    elif initial_sandbox_response is not None:
        repo = g.get_repo(repo_full_name)
        commit_hash = repo.get_commits()[0].sha
        success = initial_sandbox_response.outputs and initial_sandbox_response.success
        status = "✓" if success else "X"
        sandbox_execution_message = (
            "\n\n## GitHub Actions"
            + status
            + "\n\nHere are the GitHub Actions logs prior to making any changes:\n\n"
        )
        sandbox_execution_message += entities_create_error_logs(
            f'<a href="https://github.com/{repo_full_name}/commit/{commit_hash}"><code>{commit_hash[:7]}</code></a>',
            initial_sandbox_response,
            initial_sandbox_response_file,
        )
        if success:
            sandbox_execution_message += f"\n\nSandbox passed on the latest `{repo.default_branch}`, so sandbox checks will be enabled for this issue."
        else:
            sandbox_execution_message += "\n\nSandbox failed, so all sandbox checks will be disabled for this issue."

    if index < 0:
        index = 0
    if index == 4:
        return (
            pr_message
            + config_pr_message
            + f"\n\n---\n{user_settings.get_message(completed=True)}"
            + f"\n\n---\n{actions_message}"
            + sandbox_execution_message
        )

    total = len(progress_headers)
    index += 1 if done else 0
    index *= 100 / total
    index = int(index)
    index = min(100, index)
    if errored:
        pbar = f"\n\n<img src='https://progress-bar.dev/{index}/?&title=Errored&width=600' alt='{index}%' />"
        return (
            f"{center(sweeping_gif)}<br/>{center(pbar)}\n\n"
            + f"\n\n---\n{actions_message}"
            + sandbox_execution_message
        )
    pbar = f"\n\n<img src='https://progress-bar.dev/{index}/?&title=Progress&width=600' alt='{index}%' />"
    return (
        f"{center(sweeping_gif)}"
        + (
            center(
                f'\n\n<h2>✨ Track Sweep\'s progress on our <a href="{PROGRESS_BASE_URL}/issues/{tracking_id}">progress dashboard</a>!</h2>'
            )
            if MONGODB_URI is not None
            else ""
        )
        + f"<br/>{center(pbar)}"
        + ("\n" + stars_suffix if index != -1 else "")
        + "\n"
        + center(payment_message_start)
        + f"\n\n---\n{user_settings_message}"
        + config_pr_message
        + f"\n\n---\n{actions_message}"
        + sandbox_execution_message
    )


def on_ticket(
    title: str,
    summary: str,
    issue_number: int,
    issue_url: str, # purely for logging purposes
    username: str,
    repo_full_name: str,
    repo_description: str,
    installation_id: int,
    comment_id: int = None,
    edited: bool = False,
    tracking_id: str | None = None,
):
    with logger.contextualize(
        tracking_id=tracking_id,
    ):
        if tracking_id is None:
            tracking_id = get_hash()
        on_ticket_start_time = time()
        logger.info(f"Starting on_ticket with title {title} and summary {summary}")
        (
            title,
            slow_mode,
            do_map,
            subissues_mode,
            sandbox_mode,
            fast_mode,
            lint_mode,
        ) = strip_sweep(title)

        summary = summary or ""
        summary = re.sub(
            "<details (open)?>(\r)?\n<summary>Checklist</summary>.*",
            "",
            summary,
            flags=re.DOTALL,
        ).strip()
        summary = re.sub(
            "---\s+Checklist:(\r)?\n(\r)?\n- \[[ X]\].*",
            "",
            summary,
            flags=re.DOTALL,
        ).strip()
        summary = re.sub(
            "### Details\n\n_No response_", "", summary, flags=re.DOTALL
        )
        summary = re.sub("\n\n", "\n", summary, flags=re.DOTALL)

        repo_name = repo_full_name
        user_token, g = get_github_client(installation_id)
        repo = g.get_repo(repo_full_name)
        current_issue: Issue = repo.get_issue(number=issue_number)
        assignee = current_issue.assignee.login if current_issue.assignee else None
        if assignee is None:
            assignee = current_issue.user.login

        ticket_progress = TicketProgress(
            tracking_id=tracking_id,
            username=username,
            context=TicketContext(
                title=title,
                description=summary,
                repo_full_name=repo_full_name,
                issue_number=issue_number,
                is_public=repo.private is False,
                start_time=int(time()),
            ),
        )
        branch_match = re.search(
            r"([B|b]ranch:) *(?P<branch_name>.+?)(\s|$)", summary
        )
        overrided_branch_name = None
        if branch_match and "branch_name" in branch_match.groupdict():
            overrided_branch_name = (
                branch_match.groupdict()["branch_name"].strip().strip("`\"'")
            )
            # TODO: this code might be finicky, might have missed edge cases
            if overrided_branch_name.startswith("https://github.com/"):
                overrided_branch_name = overrided_branch_name.split("?")[0].split(
                    "tree/"
                )[-1]
            SweepConfig.get_branch(repo, overrided_branch_name)

        chat_logger = (
            ChatLogger(
                {
                    "repo_name": repo_name,
                    "title": title,
                    "summary": summary,
                    "issue_number": issue_number,
                    "issue_url": issue_url,
                    "username": (
                        username if not username.startswith("sweep") else assignee
                    ),
                    "repo_full_name": repo_full_name,
                    "repo_description": repo_description,
                    "installation_id": installation_id,
                    "type": "ticket",
                    "mode": ENV,
                    "comment_id": comment_id,
                    "edited": edited,
                    "tracking_id": tracking_id,
                },
                active=True,
            )
            if MONGODB_URI
            else None
        )

        if chat_logger and not IS_SELF_HOSTED:
            is_paying_user = chat_logger.is_paying_user()
            is_consumer_tier = chat_logger.is_consumer_tier()
            use_faster_model = chat_logger.use_faster_model()
        else:
            is_paying_user = True
            is_consumer_tier = False
            use_faster_model = False

        if use_faster_model:
            raise Exception(FASTER_MODEL_MESSAGE)

        if fast_mode:
            use_faster_model = True

        if not comment_id and not edited and chat_logger and not sandbox_mode:
            fire_and_forget_wrapper(chat_logger.add_successful_ticket)(
                gpt3=use_faster_model
            )

        organization, repo_name = repo_full_name.split("/")
        metadata = {
            "issue_url": issue_url,
            "repo_full_name": repo_full_name,
            "organization": organization,
            "repo_name": repo_name,
            "repo_description": repo_description,
            "username": username,
            "comment_id": comment_id,
            "title": title,
            "installation_id": installation_id,
            "function": "on_ticket",
            "edited": edited,
            "model": "gpt-3.5" if use_faster_model else "gpt-4",
            "tier": "pro" if is_paying_user else "free",
            "mode": ENV,
            "slow_mode": slow_mode,
            "do_map": do_map,
            "subissues_mode": subissues_mode,
            "sandbox_mode": sandbox_mode,
            "fast_mode": fast_mode,
            "is_self_hosted": IS_SELF_HOSTED,
            "tracking_id": tracking_id,
        }
        fire_and_forget_wrapper(posthog.capture)(
            username, "started", properties=metadata
        )

        try:
            if current_issue.state == "closed":
                fire_and_forget_wrapper(posthog.capture)(
                    username,
                    "issue_closed",
                    properties={
                        **metadata,
                        "duration": round(time() - on_ticket_start_time),
                    },
                )
                return {"success": False, "reason": "Issue is closed"}

            fire_and_forget_wrapper(add_emoji)(current_issue, comment_id)
            fire_and_forget_wrapper(remove_emoji)(
                current_issue, comment_id, content_to_delete="rocket"
            )
            fire_and_forget_wrapper(current_issue.edit)(body=summary)

            replies_text = ""
            summary = summary if summary else ""

            fire_and_forget_wrapper(delete_old_prs)(repo, issue_number)

            if not sandbox_mode:
                progress_headers = [
                    None,
                    "Step 1: 🔎 Searching",
                    "Step 2: ⌨️ Coding",
                    "Step 3: 🔁 Code Review",
                ]
            else:
                progress_headers = [
                    None,
                    "📖 Reading File",
                    "🛠️ Executing Sandbox",
                ]

            issue_comment = None
            payment_message, payment_message_start = get_payment_messages(
                chat_logger
            )

            ticket_progress.context.payment_context = PaymentContext(
                use_faster_model=use_faster_model,
                pro_user=is_paying_user,
                daily_tickets_used=(
                    chat_logger.get_ticket_count(use_date=True)
                    if chat_logger
                    else 0
                ),
                monthly_tickets_used=(
                    chat_logger.get_ticket_count() if chat_logger else 0
                ),
            )
            ticket_progress.save()

            config_pr_url = None
            user_settings = UserSettings.from_username(username=username)
            user_settings_message = user_settings.get_message()

            cloned_repo = ClonedRepo(
                repo_full_name,
                installation_id=installation_id,
                token=user_token,
                repo=repo,
                branch=overrided_branch_name,
            )
            # check that repo's directory is non-empty
            if os.listdir(cloned_repo.cached_dir) == []:
                logger.info("Empty repo")
                first_comment = (
                    "Sweep is currently not supported on empty repositories. Please add some"
                    f" code to your repository and try again.\n{sep}##"
                    f" {progress_headers[1]}\n{bot_suffix}{discord_suffix}"
                )
                if issue_comment is None:
                    issue_comment = current_issue.create_comment(
                        first_comment + BOT_SUFFIX
                    )
                else:
                    issue_comment.edit(first_comment + BOT_SUFFIX)
                return {"success": False}
            indexing_message = (
                "I'm searching for relevant snippets in your repository. If this is your first"
                " time using Sweep, I'm indexing your repository. You can monitor the progress using the progress dashboard"
            )
            first_comment = (
                f"{get_comment_header(0, g, repo_full_name, user_settings, progress_headers, tracking_id, payment_message_start, user_settings_message)}\n{sep}I am currently looking into this ticket! I"
                " will update the progress of the ticket in this comment. I am currently"
                f" searching through your code, looking for relevant snippets.\n{sep}##"
                f" {progress_headers[1]}\n{indexing_message}{bot_suffix}{discord_suffix}"
            )
            # Find Sweep's previous comment
            comments = []
            for comment in current_issue.get_comments():
                comments.append(comment)
                if comment.user.login == CURRENT_USERNAME:
                    issue_comment = comment
                    break
            if issue_comment is None:
                issue_comment = current_issue.create_comment(first_comment)
            else:
                fire_and_forget_wrapper(issue_comment.edit)(first_comment)
            old_edit = issue_comment.edit
            issue_comment.edit = lambda msg: old_edit(msg + BOT_SUFFIX)
            past_messages = {}
            current_index = 0
            table = None
            initial_sandbox_response = -1
            initial_sandbox_response_file = None

            def refresh_token():
                user_token, g = get_github_client(installation_id)
                repo = g.get_repo(repo_full_name)
                return user_token, g, repo

            def edit_sweep_comment(
                message: str,
                index: int,
                pr_message="",
                done=False,
                add_bonus_message=True,
            ):
                nonlocal current_index, user_token, g, repo, issue_comment, initial_sandbox_response, initial_sandbox_response_file
                # -1 = error, -2 = retry
                # Only update the progress bar if the issue generation errors.
                errored = index == -1
                if index >= 0:
                    past_messages[index] = message
                    current_index = index

                agg_message = None
                # Include progress history
                # index = -2 is reserved for
                for i in range(
                    current_index + 2
                ):  # go to next header (for Working on it... text)
                    if i == 0 or i >= len(progress_headers):
                        continue  # skip None header
                    header = progress_headers[i]
                    if header is not None:
                        header = "## " + header + "\n"
                    else:
                        header = "No header\n"
                    msg = header + (past_messages.get(i) or "Working on it...")
                    if agg_message is None:
                        agg_message = msg
                    else:
                        agg_message = agg_message + f"\n{sep}" + msg

                suffix = bot_suffix + discord_suffix
                if errored:
                    agg_message = (
                        "## ❌ Unable to Complete PR"
                        + "\n"
                        + message
                        + (
                            "\n\nFor bonus GPT-4 tickets, please report this bug on"
                            f" **[Discord](https://discord.gg/invite/sweep)** (tracking ID: `{tracking_id}`)."
                            if add_bonus_message
                            else ""
                        )
                    )
                    if table is not None:
                        agg_message = (
                            agg_message
                            + f"\n{sep}Please look at the generated plan. If something looks"
                            f" wrong, please add more details to your issue.\n\n{table}"
                        )
                    suffix = bot_suffix  # don't include discord suffix for error messages

                # Update the issue comment
                msg = f"{get_comment_header(current_index, g, repo_full_name, user_settings, progress_headers, tracking_id, payment_message_start, user_settings_message, errored=errored, pr_message=pr_message, done=done, initial_sandbox_response=initial_sandbox_response, initial_sandbox_response_file=initial_sandbox_response_file, config_pr_url=config_pr_url)}\n{sep}{agg_message}{suffix}"
                try:
                    issue_comment.edit(msg)
                except BadCredentialsException:
                    logger.error(
                        f"Bad credentials, refreshing token (tracking ID: `{tracking_id}`)"
                    )
                    user_token, g = get_github_client(installation_id)
                    repo = g.get_repo(repo_full_name)

                    issue_comment = None
                    for comment in comments:
                        if comment.user.login == CURRENT_USERNAME:
                            issue_comment = comment
                    current_issue = repo.get_issue(number=issue_number)
                    if issue_comment is None:
                        issue_comment = current_issue.create_comment(msg)
                    else:
                        issue_comment = [
                            comment
                            for comment in current_issue.get_comments()
                            if comment.user.login == CURRENT_USERNAME
                        ][0]
                        issue_comment.edit(msg)

            if use_faster_model:
                edit_sweep_comment(
                    FASTER_MODEL_MESSAGE, -1, add_bonus_message=False
                )
                posthog.capture(
                    username,
                    "ran_out_of_tickets",
                    properties={
                        **metadata,
                        "duration": round(time() - on_ticket_start_time),
                    },
                )
                return {
                    "success": False,
                    "error_message": "We deprecated supporting GPT 3.5.",
                }

            if sandbox_mode:
                handle_sandbox_mode(
                    title, repo_full_name, repo, ticket_progress, edit_sweep_comment
                )
                return {"success": True}

            if len(title + summary) < 20:
                logger.info("Issue too short")
                edit_sweep_comment(
                    (
                        f"Please add more details to your issue. I need at least 20 characters"
                        f" to generate a plan. Please join our Discord server for support (tracking_id={tracking_id})"
                    ),
                    -1,
                )
                posthog.capture(
                    username,
                    "issue_too_short",
                    properties={
                        **metadata,
                        "duration": round(time() - on_ticket_start_time),
                    },
                )
                return {"success": True}

            prs_extracted = PRReader.extract_prs(repo, summary)
            message_summary = summary
            if prs_extracted:
                message_summary += "\n\n" + prs_extracted
                edit_sweep_comment(
                    create_collapsible(
                        "I found that you mentioned the following Pull Requests that might be important:",
                        blockquote(
                            prs_extracted,
                        ),
                    ),
                    1,
                )

            try:
                # search/context manager
                logger.info("Searching for relevant snippets...")
                snippets, tree, _, repo_context_manager = fetch_relevant_files(
                    cloned_repo,
                    title,
                    message_summary,
                    replies_text,
                    username,
                    metadata,
                    on_ticket_start_time,
                    tracking_id,
                    is_paying_user,
                    is_consumer_tier,
                    issue_url,
                    chat_logger,
                    ticket_progress,
                )
                cloned_repo = repo_context_manager.cloned_repo
            except Exception as e:
                edit_sweep_comment(
                    (
                        "It looks like an issue has occurred around fetching the files."
                        " Perhaps the repo failed to initialized. If this error persists"
                        f" contact team@sweep.dev.\n\n> @{username}, editing this issue description to include more details will automatically make me relaunch. Please join our Discord server for support (tracking_id={tracking_id})"
                    ),
                    -1,
                )
                raise Exception("Failed to fetch files") from e
            _user_token, g = get_github_client(installation_id)
            user_token, g, repo = refresh_token()
            cloned_repo.token = user_token
            repo = g.get_repo(repo_full_name)
            ticket_progress.search_progress.indexing_progress = (
                ticket_progress.search_progress.indexing_total
            )
            ticket_progress.status = TicketProgressStatus.PLANNING
            ticket_progress.save()

            # Fetch git commit history
            if not repo_description:
                repo_description = "No description provided."

            message_summary += replies_text
            # removed external search as it provides no real value and only adds noise
            # external_results = ExternalSearcher.extract_summaries(message_summary)
            # if external_results:
            #     message_summary += "\n\n" + external_results

            get_documentation_dict(repo)
            docs_results = ""
            sweep_bot = construct_sweep_bot(
                repo=repo,
                repo_name=repo_name,
                issue_url=issue_url,
                repo_description=repo_description,
                title=title,
                message_summary=message_summary,
                cloned_repo=cloned_repo,
                ticket_progress=ticket_progress,
                chat_logger=chat_logger,
                snippets=snippets,
                tree=tree,
                comments=comments,
            )
            # Check repository for sweep.yml file.
            sweep_yml_exists = False
            sweep_yml_failed = False
            for content_file in repo.get_contents(""):
                if content_file.name == "sweep.yaml":
                    sweep_yml_exists = True

                    # Check if YAML is valid
                    yaml_content = content_file.decoded_content.decode("utf-8")
                    sweep_yaml_dict = {}
                    try:
                        sweep_yaml_dict = yaml.safe_load(yaml_content)
                    except Exception:
                        logger.error(f"Failed to load YAML file: {yaml_content}")
                    if len(sweep_yaml_dict) > 0:
                        break
                    linter_config = yamllint_config.YamlLintConfig(custom_config)
                    problems = list(linter.run(yaml_content, linter_config))
                    if problems:
                        errors = [
                            f"Line {problem.line}: {problem.desc} (rule: {problem.rule})"
                            for problem in problems
                        ]
                        error_message = "\n".join(errors)
                        markdown_error_message = f"**There is something wrong with your [sweep.yaml](https://github.com/{repo_full_name}/blob/main/sweep.yaml):**\n```\n{error_message}\n```"
                        sweep_yml_failed = True
                        logger.error(markdown_error_message)
                        edit_sweep_comment(markdown_error_message, -1)
                    else:
                        logger.info("The YAML file is valid. No errors found.")
                    break

            # If sweep.yaml does not exist, then create a new PR that simply creates the sweep.yaml file.
            if not sweep_yml_exists:
                try:
                    logger.info("Creating sweep.yaml file...")
                    config_pr = create_config_pr(sweep_bot, cloned_repo=cloned_repo)
                    config_pr_url = config_pr.html_url
                    edit_sweep_comment(message="", index=-2)
                except SystemExit:
                    raise SystemExit
                except Exception as e:
                    logger.error(
                        "Failed to create new branch for sweep.yaml file.\n",
                        e,
                        traceback.format_exc(),
                    )
            else:
                logger.info("sweep.yaml file already exists.")

            try:
                # ANALYZE SNIPPETS
                newline = "\n"
                edit_sweep_comment(
                    "I found the following snippets in your repository. I will now analyze"
                    " these snippets and come up with a plan."
                    + "\n\n"
                    + create_collapsible(
                        "Some code snippets I think are relevant in decreasing order of relevance (click to expand). If some file is missing from here, you can mention the path in the ticket description.",
                        "\n".join(
                            [
                                f"https://github.com/{organization}/{repo_name}/blob/{repo.get_commits()[0].sha}/{snippet.file_path}#L{max(snippet.start, 1)}-L{min(snippet.end, snippet.content.count(newline) - 1)}\n"
                                for snippet in snippets
                            ]
                        ),
                    )
                    + (
                        create_collapsible(
                            "I also found that you mentioned the following Pull Requests that may be helpful:",
                            blockquote(prs_extracted),
                        )
                        if prs_extracted
                        else ""
                    )
                    + (f"\n\n{docs_results}\n\n" if docs_results else ""),
                    1,
                )

                logger.info("Fetching files to modify/create...")
                file_change_requests, plan = get_files_to_change(
                    relevant_snippets=repo_context_manager.current_top_snippets,
                    read_only_snippets=repo_context_manager.read_only_snippets,
                    problem_statement=f"{title}\n\n{summary}",
                    repo_name=repo_full_name,
                )
                validate_file_change_requests(file_change_requests, cloned_repo)
                ticket_progress.planning_progress.file_change_requests = (
                    file_change_requests
                )
                ticket_progress.coding_progress.file_change_requests = (
                    file_change_requests
                )
                ticket_progress.coding_progress.assistant_conversations = [
                    AssistantConversation() for fcr in file_change_requests
                ]
                ticket_progress.status = TicketProgressStatus.CODING
                ticket_progress.save()

                if not file_change_requests:
                    if len(title + summary) < 60:
                        edit_sweep_comment(
                            (
                                "Sorry, I could not find any files to modify, can you please"
                                " provide more details? Please make sure that the title and"
                                " summary of the issue are at least 60 characters."
                            ),
                            -1,
                        )
                    else:
                        edit_sweep_comment(
                            (
                                "Sorry, I could not find any files to modify, can you please"
                                " provide more details?"
                            ),
                            -1,
                        )
                    raise Exception("No files to modify.")

                file_change_requests: list[
                    FileChangeRequest
                ] = sweep_bot.validate_file_change_requests(
                    file_change_requests,
                )
                ticket_progress.planning_progress.file_change_requests = (
                    file_change_requests
                )
                ticket_progress.coding_progress.assistant_conversations = [
                    AssistantConversation() for fcr in file_change_requests
                ]
                ticket_progress.save()

                table = tabulate(
                    [
                        [
                            file_change_request.entity_display,
                            file_change_request.instructions_display.replace(
                                "\n", "<br/>"
                            ).replace("```", "\\```"),
                        ]
                        for file_change_request in file_change_requests
                        if file_change_request.change_type != "check"
                    ],
                    headers=["File Path", "Proposed Changes"],
                    tablefmt="pipe",
                )

                logger.info("Generating PR...")
                pull_request = PullRequest(
                    title="Sweep: " + title,
                    branch_name="sweep/" + to_branch_name(title),
                    content="",
                )
                logger.info("Making PR...")

                ticket_progress.context.branch_name = pull_request.branch_name
                ticket_progress.save()

                files_progress: list[tuple[str, str, str, str]] = [
                    (
                        file_change_request.entity_display,
                        file_change_request.instructions_display,
                        "⏳ In Progress",
                        "",
                    )
                    for file_change_request in file_change_requests
                ]

                checkboxes_progress: list[tuple[str, str, str]] = [
                    (
                        file_change_request.entity_display,
                        file_change_request.instructions_display,
                        " ",
                    )
                    for file_change_request in file_change_requests
                    if not file_change_request.change_type == "check"
                ]
                checkboxes_contents = "\n".join(
                    [
                        create_checkbox(
                            f"`{filename}`", blockquote(instructions), check == "X"
                        )
                        for filename, instructions, check in checkboxes_progress
                    ]
                )
                create_collapsible("Checklist", checkboxes_contents, opened=True)

                file_change_requests[0].status = "running"

                condensed_checkboxes_contents = "\n".join(
                    [
                        create_checkbox(f"`{filename}`", "", check == "X").strip()
                        for filename, instructions, check in checkboxes_progress
                    ]
                )
                condensed_checkboxes_collapsible = create_collapsible(
                    "Checklist", condensed_checkboxes_contents, opened=True
                )

                current_issue = repo.get_issue(number=issue_number)
                current_issue.edit(
                    body=summary + "\n\n" + condensed_checkboxes_collapsible
                )

                delete_branch = False

                generator = create_pr_changes(
                    file_change_requests,
                    pull_request,
                    sweep_bot,
                    username,
                    installation_id,
                    issue_number,
                    chat_logger=chat_logger,
                    base_branch=overrided_branch_name,
                    additional_messages=[],
                )
                edit_sweep_comment(checkboxes_contents, 2)
                if not file_change_requests:
                    raise NoFilesException()
                response = {
                    "error": Exception(
                        f"Sweep failed to generate any file change requests! This could mean that Sweep failed to find the correct lines of code to modify or that GPT-4 did not respond in our specified format. Sometimes, retrying will fix this error. Otherwise, reach out to our Discord server for support (tracking_id={tracking_id})."
                    )
                }

                changed_files = []
                for item in generator:
                    if isinstance(item, dict):
                        response = item
                        break
                    (
                        new_file_contents,
                        _,
                        commit,
                        file_change_requests,
                    ) = item
                    # append all files that have been changed
                    if new_file_contents:
                        for file_name, _ in new_file_contents.items():
                            changed_files.append(file_name)
                    commit_hash: str = (
                        commit
                        if isinstance(commit, str)
                        else (
                            commit.sha
                            if commit is not None
                            else repo.get_branch(
                                pull_request.branch_name
                            ).commit.sha
                        )
                    )
                    commit_url = (
                        f"https://github.com/{repo_full_name}/commit/{commit_hash}"
                    )
                    commit_url_display = (
                        f"<a href='{commit_url}'><code>{commit_hash[:7]}</code></a>"
                    )
                    create_error_logs(
                        commit_url_display,
                        None,
                        status=(
                            "✓"
                        ),
                    )
                    checkboxes_progress = [
                        (
                            file_change_request.display_summary
                            + " "
                            + file_change_request.status_display
                            + " "
                            + (file_change_request.commit_hash_url or "")
                            + f" [Edit]({file_change_request.get_edit_url(repo.full_name, pull_request.branch_name)})",
                            file_change_request.instructions_ticket_display
                            + f"\n\n{file_change_request.diff_display}",
                            (
                                "X"
                                if file_change_request.status
                                in ("succeeded", "failed")
                                else " "
                            ),
                        )
                        for file_change_request in file_change_requests
                    ]
                    checkboxes_contents = "\n".join(
                        [
                            checkbox_template.format(
                                check=check,
                                filename=filename,
                                instructions=blockquote(instructions),
                            )
                            for filename, instructions, check in checkboxes_progress
                        ]
                    )
                    collapsible_template.format(
                        summary="Checklist",
                        body=checkboxes_contents,
                        opened="open",
                    )
                    condensed_checkboxes_contents = "\n".join(
                        [
                            checkbox_template.format(
                                check=check,
                                filename=filename,
                                instructions="",
                            ).strip()
                            for filename, instructions, check in checkboxes_progress
                            if not instructions.lower().startswith("run")
                        ]
                    )
                    condensed_checkboxes_collapsible = collapsible_template.format(
                        summary="Checklist",
                        body=condensed_checkboxes_contents,
                        opened="open",
                    )

                    try:
                        current_issue = repo.get_issue(number=issue_number)
                    except BadCredentialsException:
                        user_token, g, repo = refresh_token()
                        cloned_repo.token = user_token

                    current_issue.edit(
                        body=summary + "\n\n" + condensed_checkboxes_collapsible
                    )

                    logger.info(files_progress)
                    edit_sweep_comment(checkboxes_contents, 2)
                if not response.get("success"):
                    raise Exception(f"Failed to create PR: {response.get('error')}")

                checkboxes_contents = "\n".join(
                    [
                        checkbox_template.format(
                            check=check,
                            filename=filename,
                            instructions=blockquote(instructions),
                        )
                        for filename, instructions, check in checkboxes_progress
                    ]
                )
                condensed_checkboxes_contents = "\n".join(
                    [
                        checkbox_template.format(
                            check=check,
                            filename=filename,
                            instructions="",
                        ).strip()
                        for filename, instructions, check in checkboxes_progress
                        if not instructions.lower().startswith("run")
                    ]
                )
                condensed_checkboxes_collapsible = collapsible_template.format(
                    summary="Checklist",
                    body=condensed_checkboxes_contents,
                    opened="open",
                )
                for _ in range(3):
                    try:
                        current_issue.edit(
                            body=summary + "\n\n" + condensed_checkboxes_collapsible
                        )
                        break
                    except Exception:
                        from time import sleep

                        sleep(1)
                edit_sweep_comment(checkboxes_contents, 2)

                pr_changes = response["pull_request"]
                # change the body here
                diff_text = get_branch_diff_text(
                    repo=repo,
                    branch=pull_request.branch_name,
                    base_branch=overrided_branch_name,
                )
                new_description = PRDescriptionBot().describe_diffs(
                    diff_text,
                    pull_request.title,
                )
                # TODO: update the title as well
                if new_description:
                    pr_changes.body = (
                        f"{new_description}\n\nFixes"
                        f" #{issue_number}.\n\n---\n\n{UPDATES_MESSAGE}\n\n---\n\n{INSTRUCTIONS_FOR_REVIEW}{BOT_SUFFIX}"
                    )

                edit_sweep_comment(
                    "I have finished coding the issue. I am now reviewing it for completeness.",
                    3,
                )
                change_location = f" [`{pr_changes.pr_head}`](https://github.com/{repo_full_name}/commits/{pr_changes.pr_head}).\n\n"
                review_message = (
                    "Here are my self-reviews of my changes at" + change_location
                )

                try:
                    fire_and_forget_wrapper(remove_emoji)(content_to_delete="eyes")
                except SystemExit:
                    raise SystemExit
                except Exception:
                    pass

                changes_required, review_message = False, ""
                if changes_required:
                    edit_sweep_comment(
                        review_message
                        + "\n\nI finished incorporating these changes.",
                        3,
                    )
                else:
                    edit_sweep_comment(
                        f"I have finished reviewing the code for completeness. I did not find errors for {change_location}",
                        3,
                    )

                pr_actions_message = (
                    create_action_buttons(
                        [
                            SWEEP_GOOD_FEEDBACK,
                            SWEEP_BAD_FEEDBACK,
                        ],
                        header="### PR Feedback (click)\n",
                    )
                    + "\n"
                    if DISCORD_FEEDBACK_WEBHOOK_URL is not None
                    else ""
                )
                revert_buttons = []
                for changed_file in set(changed_files):
                    revert_buttons.append(
                        Button(label=f"{RESET_FILE} {changed_file}")
                    )
                revert_buttons_list = ButtonList(
                    buttons=revert_buttons, title=REVERT_CHANGED_FILES_TITLE
                )

                rule_buttons = []
                repo_rules = get_rules(repo) or []
                if repo_rules != [""] and repo_rules != []:
                    for rule in repo_rules or []:
                        if rule:
                            rule_buttons.append(
                                Button(label=f"{RULES_LABEL} {rule}")
                            )
                    if len(repo_rules) == 0:
                        for rule in DEFAULT_RULES:
                            rule_buttons.append(
                                Button(label=f"{RULES_LABEL} {rule}")
                            )

                rules_buttons_list = ButtonList(
                    buttons=rule_buttons, title=RULES_TITLE
                )

                sandbox_passed = None
                for file_change_request in file_change_requests:
                    if file_change_request.change_type == "check":
                        if (
                            file_change_request.sandbox_response
                            and file_change_request.sandbox_response.error_messages
                        ):
                            sandbox_passed = False
                        elif sandbox_passed is None:
                            sandbox_passed = True

                # delete failing sweep yaml if applicable
                if sweep_yml_failed:
                    try:
                        repo.delete_file(
                            "sweep.yaml",
                            "Delete failing sweep.yaml",
                            branch=pr_changes.pr_head,
                            sha=repo.get_contents("sweep.yaml").sha,
                        )
                    except Exception:
                        pass

                # create draft pr, then convert to regular pr later
                pr: GithubPullRequest = repo.create_pull(
                    title=pr_changes.title,
                    body=pr_actions_message + pr_changes.body,
                    head=pr_changes.pr_head,
                    base=overrided_branch_name or SweepConfig.get_branch(repo),
                    # TODO: reenable it later
                    draft=True,
                )

                try:
                    pr.add_to_assignees(username)
                except Exception as e:
                    logger.error(
                        f"Failed to add assignee {username}: {e}, probably a bot."
                    )

                ticket_progress.status = TicketProgressStatus.COMPLETE
                ticket_progress.context.done_time = time()
                ticket_progress.context.pr_id = pr.number
                ticket_progress.save()

                if revert_buttons:
                    pr.create_issue_comment(
                        revert_buttons_list.serialize() + BOT_SUFFIX
                    )
                if rule_buttons:
                    pr.create_issue_comment(
                        rules_buttons_list.serialize() + BOT_SUFFIX
                    )

                # add comments before labelling
                pr.add_to_labels(GITHUB_LABEL_NAME)
                current_issue.create_reaction("rocket")
                heres_pr_message = f'<h1 align="center">🚀 Here\'s the PR! <a href="{pr.html_url}">#{pr.number}</a></h1>'
                progress_message = f'<div align="center"><b>See Sweep\'s progress at <a href="{PROGRESS_BASE_URL}/issues/{tracking_id}">the progress dashboard</a>!</b></div>'
                edit_sweep_comment(
                    review_message + "\n\nSuccess! 🚀",
                    4,
                    pr_message=(
                        f"{center(heres_pr_message)}\n{center(progress_message)}\n{center(payment_message_start)}"
                    ),
                    done=True,
                )

                user_settings = UserSettings.from_username(username=username)
                user = g.get_user(username)
                full_name = user.name or user.login
                name = full_name.split(" ")[0]
                files_changed = []
                for fcr in file_change_requests:
                    if fcr.change_type in ("create", "modify"):
                        diff = list(
                            difflib.unified_diff(
                                (fcr.old_content or "").splitlines() or [],
                                (fcr.new_content or "").splitlines() or [],
                                lineterm="",
                            )
                        )
                        added = sum(
                            1
                            for line in diff
                            if line.startswith("+") and not line.startswith("+++")
                        )
                        removed = sum(
                            1
                            for line in diff
                            if line.startswith("-") and not line.startswith("---")
                        )
                        files_changed.append(
                            f"<code>{fcr.filename}</code> (+{added}/-{removed})"
                        )
                user_settings.send_email(
                    subject=f"Sweep Pull Request Complete for {repo_name}#{issue_number} {title}",
                    html=email_template.format(
                        name=name,
                        pr_url=pr.html_url,
                        issue_number=issue_number,
                        repo_full_name=repo_full_name,
                        pr_number=pr.number,
                        progress_url=f"{PROGRESS_BASE_URL}/issues/{tracking_id}",
                        summary=markdown.markdown(pr_changes.body),
                        files_changed="\n".join(
                            [f"<li>{item}</li>" for item in files_changed]
                        ),
                        sweeping_gif=sweeping_gif,
                    ),
                )

                # poll for github to check when gha are done
                total_poll_attempts = 0
                total_edit_attempts = 0
                SLEEP_DURATION_SECONDS = 15
                GITHUB_ACTIONS_ENABLED = get_gha_enabled(repo=repo) and DEPLOYMENT_GHA_ENABLED
                GHA_MAX_EDIT_ATTEMPTS = 1 # max number of times to edit PR
                while True and GITHUB_ACTIONS_ENABLED:
                    logger.info(
                        f"Polling to see if Github Actions have finished... {total_poll_attempts}"
                    )
                    # we wait at most 60 minutes
                    if total_poll_attempts * SLEEP_DURATION_SECONDS // 60 >= 60:
                        break
                    else:
                        # wait one minute between check attempts
                        total_poll_attempts += 1
                        from time import sleep

                        sleep(SLEEP_DURATION_SECONDS)
                    runs = list(repo.get_workflow_runs(branch=pr.head.ref, head_sha=pr.head.sha))
                    # if all runs have succeeded, break
                    if all([run.conclusion == "success" for run in runs]):
                        break
                    # if any of them have failed we retry
                    if any([run.conclusion == "failure" for run in runs]):
                        failed_runs = [
                            run for run in runs if run.conclusion == "failure"
                        ]

                        failed_gha_logs: list[str] = get_failing_gha_logs(
                            failed_runs,
                            installation_id,
                        )
                        if failed_gha_logs:
                            # make edits to the PR
                            # TODO: look into rollbacks so we don't continue adding onto errors
                            cloned_repo = ClonedRepo( # reinitialize cloned_repo to avoid conflicts
                                repo_full_name,
                                installation_id=installation_id,
                                token=user_token,
                                repo=repo,
                                branch=pr.head.ref,
                            )
                            diffs = get_branch_diff_text(repo=repo, branch=pr.head.ref, base_branch=pr.base.ref)
                            problem_statement = f"{title}\n{summary}\n{replies_text}"
                            all_information_prompt = f"While trying to address the user request:\n<user_request>\n{problem_statement}\n</user_request>\n{failed_gha_logs}\nThese are the changes that were previously made:\n<diffs>\n{diffs}\n</diffs>\n\nFix the failing logs."
                            
                            repo_context_manager = prep_snippets(cloned_repo=cloned_repo, query=(title + summary + replies_text).strip("\n"), ticket_progress=ticket_progress) # need to do this, can use the old query for speed
                            repo_context_manager = get_relevant_context(
                                all_information_prompt,
                                repo_context_manager,
                                ticket_progress,
                                chat_logger=chat_logger,
                                import_graph=None,
                                num_rollouts=1,
                            )
                            sweep_bot: SweepBot = construct_sweep_bot(
                                repo=repo,
                                repo_name=repo_name,
                                issue_url=issue_url,
                                repo_description=repo_description,
                                title="Fix the following errors to complete the user request.",
                                message_summary=all_information_prompt,
                                cloned_repo=cloned_repo,
                                ticket_progress=ticket_progress,
                                chat_logger=chat_logger,
                                snippets=snippets,
                                tree=tree,
                                comments=comments,
                            )
                            file_change_requests, plan = get_files_to_change(
                                relevant_snippets=repo_context_manager.current_top_snippets,
                                read_only_snippets=repo_context_manager.read_only_snippets,
                                problem_statement=all_information_prompt,
                                repo_name=repo_full_name,
                            )
                            validate_file_change_requests(file_change_requests, cloned_repo)
                            previous_modify_files_dict: dict[str, dict[str, str | list[str]]] | None = None
                            sweep_bot.handle_modify_file_main(
                                branch=pr.head.ref,
                                assistant_conversation=None,
                                additional_messages=[],
                                previous_modify_files_dict=previous_modify_files_dict,
                                file_change_requests=file_change_requests
                            )
                            pr = repo.get_pull(pr.number) # IMPORTANT: resync PR otherwise you'll fetch old GHA runs
                            total_edit_attempts += 1
                            if total_edit_attempts >= GHA_MAX_EDIT_ATTEMPTS:
                                logger.info(f"Tried to edit PR {GHA_MAX_EDIT_ATTEMPTS} times, giving up.")
                                break
                    # if none of the runs have completed we wait and poll github
                    logger.info(
                        f"No Github Actions have failed yet and not all have succeeded yet, waiting for {SLEEP_DURATION_SECONDS} seconds before polling again..."
                    )
                # break from main for loop
                convert_pr_draft_field(pr, is_draft=False, installation_id=installation_id)
            except MaxTokensExceeded as e:
                logger.info("Max tokens exceeded")
                ticket_progress.status = TicketProgressStatus.ERROR
                ticket_progress.error_message = "Max tokens exceeded. Feel free to add more details to the issue descript for Sweep to better address it, or alternatively, reach out to Kevin or William for help at https://discord.gg/sweep."
                ticket_progress.save()
                log_error(
                    is_paying_user,
                    is_consumer_tier,
                    username,
                    issue_url,
                    "Max Tokens Exceeded",
                    str(e) + "\n" + traceback.format_exc(),
                    priority=2,
                )
                if chat_logger and chat_logger.is_paying_user():
                    edit_sweep_comment(
                        (
                            f"Sorry, I could not edit `{e.filename}` as this file is too long."
                            " We are currently working on improved file streaming to address"
                            " this issue.\n"
                        ),
                        -1,
                    )
                else:
                    edit_sweep_comment(
                        (
                            f"Sorry, I could not edit `{e.filename}` as this file is too"
                            " long.\n\nIf this file is incorrect, please describe the desired"
                            " file in the prompt. However, if you would like to edit longer"
                            " files, consider upgrading to [Sweep Pro](https://sweep.dev/) for"
                            " longer context lengths.\n"
                        ),
                        -1,
                    )
                delete_branch = True
                raise e
            except NoFilesException as e:
                ticket_progress.status = TicketProgressStatus.ERROR
                ticket_progress.error_message = "Sweep could not find files to modify to address this issue. Feel free to add more details to the issue descript for Sweep to better address it, or alternatively, reach out to Kevin or William for help at https://discord.gg/sweep."
                ticket_progress.save()

                logger.info("Sweep could not find files to modify")
                log_error(
                    is_paying_user,
                    is_consumer_tier,
                    username,
                    issue_url,
                    "Sweep could not find files to modify",
                    str(e) + "\n" + traceback.format_exc(),
                    priority=2,
                )
                edit_sweep_comment(
                    (
                        "Sorry, Sweep could not find any appropriate files to edit to address"
                        " this issue. If this is a mistake, please provide more context and Sweep"
                        f" will retry!\n\n> @{username}, please edit the issue description to"
                        " include more details about this issue."
                    ),
                    -1,
                )
                delete_branch = True
                raise e
            except openai.BadRequestError as e:
                ticket_progress.status = TicketProgressStatus.ERROR
                ticket_progress.error_message = "Sorry, it looks like there is an error with communicating with OpenAI. If this error persists, reach out to Kevin or William for help at https://discord.gg/sweep."
                ticket_progress.save()

                logger.error(traceback.format_exc())
                logger.error(e)
                edit_sweep_comment(
                    (
                        "I'm sorry, but it looks our model has ran out of context length. We're"
                        " trying to make this happen less, but one way to mitigate this is to"
                        " code smaller files. If this error persists report it at"
                        " https://discord.gg/sweep."
                    ),
                    -1,
                )
                log_error(
                    is_paying_user,
                    is_consumer_tier,
                    username,
                    issue_url,
                    "Context Length",
                    str(e) + "\n" + traceback.format_exc(),
                    priority=2,
                )
                posthog.capture(
                    username,
                    "failed",
                    properties={
                        "error": str(e),
                        "trace": traceback.format_exc(),
                        "reason": "Invalid request error / context length",
                        **metadata,
                        "duration": round(time() - on_ticket_start_time),
                    },
                )
                delete_branch = True
                raise e
            except AssistantRaisedException as e:
                if ticket_progress is not None:
                    ticket_progress.status = TicketProgressStatus.ERROR
                    ticket_progress.error_message = f"Sweep raised an error with the following message: {e.message}. Feel free to add more details to the issue descript for Sweep to better address it, or alternatively, reach out to Kevin or William for help at https://discord.gg/sweep."
                    ticket_progress.save()

                logger.exception(e)
                edit_sweep_comment(
                    f"Sweep raised an error with the following message:\n{blockquote(e.message)}",
                    -1,
                )
                log_error(
                    is_paying_user,
                    is_consumer_tier,
                    username,
                    issue_url,
                    "Workflow",
                    str(e) + "\n" + traceback.format_exc(),
                    priority=1,
                )
                raise e
            except Exception as e:
                ticket_progress.status = TicketProgressStatus.ERROR
                ticket_progress.error_message = f"Internal server error: {str(e)}. Feel free to add more details to the issue descript for Sweep to better address it, or alternatively, reach out to Kevin or William for help at https://discord.gg/sweep."
                ticket_progress.save()

                logger.error(traceback.format_exc())
                logger.error(e)
                # title and summary are defined elsewhere
                if len(title + summary) < 60:
                    edit_sweep_comment(
                        (
                            "I'm sorry, but it looks like an error has occurred due to"
                            + " a planning failure. Feel free to add more details to the issue description"
                            + " so Sweep can better address it. Alternatively, reach out to Kevin or William for help at"
                            + " https://discord.gg/sweep."
                        ),
                        -1,
                    )
                else:
                    edit_sweep_comment(
                        (
                            "I'm sorry, but it looks like an error has occurred due to"
                            + " a planning failure. Feel free to add more details to the issue description"
                            + " so Sweep can better address it. Alternatively, reach out to Kevin or William for help at"
                            + " https://discord.gg/sweep."
                        ),
                        -1,
                    )
                log_error(
                    is_paying_user,
                    is_consumer_tier,
                    username,
                    issue_url,
                    "Workflow",
                    str(e) + "\n" + traceback.format_exc(),
                    priority=1,
                )
                raise e
            else:
                try:
                    fire_and_forget_wrapper(remove_emoji)(content_to_delete="eyes")
                    fire_and_forget_wrapper(add_emoji)("rocket")
                except SystemExit:
                    raise SystemExit
                except Exception as e:
                    logger.error(e)

            if delete_branch:
                try:
                    if pull_request.branch_name.startswith("sweep"):
                        repo.get_git_ref(
                            f"heads/{pull_request.branch_name}"
                        ).delete()
                    else:
                        raise Exception(
                            f"Branch name {pull_request.branch_name} does not start with sweep/"
                        )
                except SystemExit:
                    raise SystemExit
                except Exception as e:
                    logger.error(e)
                    logger.error(traceback.format_exc())
                    logger.info("Deleted branch", pull_request.branch_name)
        except Exception as e:
            posthog.capture(
                username,
                "failed",
                properties={
                    **metadata,
                    "error": str(e),
                    "trace": traceback.format_exc(),
                    "duration": round(time() - on_ticket_start_time),
                },
            )
            raise e
        posthog.capture(
            username,
            "success",
            properties={**metadata, "duration": round(time() - on_ticket_start_time)},
        )
        logger.info("on_ticket success in " + str(round(time() - on_ticket_start_time)))
        return {"success": True}


def handle_sandbox_mode(
    title, repo_full_name, repo, ticket_progress, edit_sweep_comment
):
    logger.info("Running in sandbox mode")
    sweep_bot = SweepBot(repo=repo, ticket_progress=ticket_progress)
    logger.info("Getting file contents")
    file_name = title.split(":")[1].strip()
    file_contents = sweep_bot.get_contents(file_name).decoded_content.decode("utf-8")
    try:
        ext = file_name.split(".")[-1]
    except Exception:
        ext = ""
    file_contents.replace("```", "\`\`\`")
    sha = repo.get_branch(repo.default_branch).commit.sha
    permalink = f"https://github.com/{repo_full_name}/blob/{sha}/{file_name}#L1-L{len(file_contents.splitlines())}"
    logger.info("Running sandbox")
    edit_sweep_comment(
        f"Running sandbox for {file_name}. Current Code:\n\n{permalink}",
        1,
    )
    updated_contents, sandbox_response = sweep_bot.check_sandbox(
        file_name, file_contents
    )
    logger.info("Sandbox finished")
    logs = (
        (
            "<br/>"
            + create_collapsible(
                "Sandbox logs",
                blockquote(
                    "\n\n".join(
                        [
                            create_collapsible(
                                f"<code>{output}</code> {i + 1}/{len(sandbox_response.outputs)} {format_sandbox_success(sandbox_response.success)}",
                                f"<pre>{clean_logs(output)}</pre>",
                                i == len(sandbox_response.outputs) - 1,
                            )
                            for i, output in enumerate(sandbox_response.outputs)
                            if len(sandbox_response.outputs) > 0
                        ]
                    )
                ),
                opened=True,
            )
        )
        if sandbox_response
        else ""
    )

    updated_contents = updated_contents.replace("```", "\`\`\`")
    diff = generate_diff(file_contents, updated_contents).replace("```", "\`\`\`")
    diff_display = (
        f"Updated Code:\n\n```{ext}\n{updated_contents}```\nDiff:\n```diff\n{diff}\n```"
        if diff
        else f"Sandbox made no changes to {file_name} (formatters were not configured or Sweep didn't make changes)."
    )

    edit_sweep_comment(
        f"{logs}\n{diff_display}",
        2,
    )
    edit_sweep_comment("N/A", 3)
    logger.info("Sandbox comments updated")


def get_branch_diff_text(repo, branch, base_branch=None):
    base_branch = base_branch or SweepConfig.get_branch(repo)
    comparison = repo.compare(base_branch, branch)
    file_diffs = comparison.files

    pr_diffs = []
    for file in file_diffs:
        diff = file.patch
        if (
            file.status == "added"
            or file.status == "modified"
            or file.status == "removed"
        ):
            pr_diffs.append((file.filename, diff))
        else:
            logger.info(
                f"File status {file.status} not recognized"
            )  # TODO(sweep): We don't handle renamed files
    return "\n".join([f"{filename}\n{diff}" for filename, diff in pr_diffs])


def get_payment_messages(chat_logger: ChatLogger):
    if chat_logger:
        is_paying_user = chat_logger.is_paying_user()
        is_consumer_tier = chat_logger.is_consumer_tier()
        use_faster_model = chat_logger.use_faster_model()
    else:
        is_paying_user = True
        is_consumer_tier = False
        use_faster_model = False

    tracking_id = chat_logger.data["tracking_id"] if MONGODB_URI is not None else None

    # Find the first comment made by the bot
    tickets_allocated = 5
    if is_consumer_tier:
        tickets_allocated = 15
    if is_paying_user:
        tickets_allocated = 500
    purchased_ticket_count = (
        chat_logger.get_ticket_count(purchased=True) if chat_logger else 0
    )
    ticket_count = (
        max(tickets_allocated - chat_logger.get_ticket_count(), 0)
        + purchased_ticket_count
        if chat_logger
        else 999
    )
    daily_ticket_count = (
        (3 - chat_logger.get_ticket_count(use_date=True) if not use_faster_model else 0)
        if chat_logger
        else 999
    )

    model_name = "GPT-4"
    single_payment_link = "https://buy.stripe.com/00g3fh7qF85q0AE14d"
    pro_payment_link = "https://buy.stripe.com/00g5npeT71H2gzCfZ8"
    daily_message = (
        f" and {daily_ticket_count} for the day"
        if not is_paying_user and not is_consumer_tier
        else ""
    )
    user_type = "💎 <b>Sweep Pro</b>" if is_paying_user else "⚡ <b>Sweep Basic Tier</b>"
    gpt_tickets_left_message = (
        f"{ticket_count} GPT-4 tickets left for the month"
        if not is_paying_user
        else "unlimited GPT-4 tickets"
    )
    purchase_message = f"<br/><br/> For more GPT-4 tickets, visit <a href={single_payment_link}>our payment portal</a>. For a one week free trial, try <a href={pro_payment_link}>Sweep Pro</a> (unlimited GPT-4 tickets)."
    payment_message = (
        f"{user_type}: I used {model_name} to create this ticket. You have {gpt_tickets_left_message}{daily_message}. (tracking ID: <code>{tracking_id}</code>)"
        + (purchase_message if not is_paying_user else "")
    )
    payment_message_start = (
        f"{user_type}: I'm using {model_name}. You have {gpt_tickets_left_message}{daily_message}. (tracking ID: <code>{tracking_id}</code>)"
        + (purchase_message if not is_paying_user else "")
    )

    return payment_message, payment_message_start


import re
from itertools import islice

from github.Repository import Repository
from pydantic import BaseModel

from loguru import logger
from sweepai.config.client import SweepConfig
from sweepai.utils.chat_logger import discord_log_error

summary_format_brief = """# Pull Request #{id_}
## Title: {pr_title}
## Files changed:

{files_touched}
"""

summary_format = """# Pull Request #{id_}

## Title: {pr_title}
## Summary:
{pr_summary}

## Here is the diff of the Pull Request:

{diff}
"""

diff_format = """Diffs for file {file_path}:
```diff
{diff}
```
"""


class PRReader(BaseModel):
    repo: Repository

    class Config:
        arbitrary_types_allowed = True

    @staticmethod
    def extract_pr_ids(content: str) -> list[str]:
        pattern_1 = r" #(?P<pr_id>\d+)"
        pattern_2 = (
            r"https://github\.com/[a-zA-Z0-9_-]+/[a-zA-Z0-9_-]+/pull/(?P<pr_id>\d+)"
        )
        return [
            _match.group("pr_id") for _match in set(re.finditer(pattern_1, content))
        ] + [_match.group("pr_id") for _match in set(re.finditer(pattern_2, content))]

    def extract_summary_from_pr_id(self, pr_id: int, sweep_config: SweepConfig = None) -> str:
        pr = self.repo.get_pull(int(pr_id))
        diff = ""
        files = list(islice(pr.get_files(), 51))
        for file in files:
            path = file.filename
            # make sure changes are within allowed directories
            if sweep_config and sweep_config.is_file_excluded(path):
                continue
            patch = file.patch
            diff += diff_format.format(file_path=path, diff=patch)
        return summary_format.format(
            id_=pr_id, pr_title=pr.title, pr_summary=pr.body, diff=diff
        )
    
    # output a truncated version of the pr summary
    def extract_summary_from_pr_id_brief(self, pr_id: int, sweep_config: SweepConfig = None) -> str:
        pr = self.repo.get_pull(int(pr_id))
        files_touched = ""
        files = list(islice(pr.get_files(), 101))
        for file in files:
            path = file.filename
            # make sure changes are within allowed directories
            if sweep_config and sweep_config.is_file_excluded(path):
                continue
            files_touched += f"{path}\n"
        return summary_format_brief.format(
            id_=pr_id, pr_title=pr.title, files_touched=files_touched
        )

    @staticmethod
    def extract_prs(repo: Repository, content: str):
        sweep_config = SweepConfig()
        logger.info("Extracting pull requests from content")
        try:
            pr_reader = PRReader(repo=repo)
            summaries = []
            result = ""
            for pr_id in pr_reader.extract_pr_ids(content):
                summaries.append(pr_reader.extract_summary_from_pr_id(pr_id, sweep_config = sweep_config))
                result += summaries[-1]
            if result:
                result = (
                    "The following PRs were mentioned in the issue:\n\n"
                    + result
                    + "\nBe sure to follow the PRs as a reference when making code changes. If the user instructs you to follow the referenced PR, limit the scope of your changes to the referenced PR."
                )
            
            # output brief version
            if len(result) > sweep_config.truncation_cutoff:
                brief_summaries = []
                result = ""
                for pr_id in pr_reader.extract_pr_ids(content):
                    brief_summaries.append(pr_reader.extract_summary_from_pr_id_brief(pr_id, sweep_config = sweep_config))
                    result += brief_summaries[-1]
                if result:
                    result = (
                        "The following PRs were mentioned in the issue:\n\n"
                        + result
                        + "\nBe sure to follow the PRs as a reference when making code changes. If the user instructs you to follow the referenced PR, limit the scope of your changes to the referenced PR."
                    )

            return result[:sweep_config.max_github_comment_body_length] # enforce hard cut off
        except Exception as e:
            logger.error(f"Failed to extract PRs from content: {e}")
            discord_log_error(f"Failed to extract PRs from content: {e}")
            return ""



import ast
import re

import tree_sitter_languages
from pydantic import BaseModel
from tree_sitter import Node, Parser, Tree


class CodeTree(BaseModel):
    code: str
    language: str
    tree: Tree

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def from_code(cls, code: str, language: str = "python"):
        parser = Parser()
        parser.set_language(tree_sitter_languages.get_language(language))
        tree = parser.parse(bytes(code, "utf8"))
        return cls(code=code, language=language, tree=tree)

    def get_path_to_line(self, min_line: int, max_line: int = -1) -> list[Node]:
        if max_line == -1:
            max_line = min_line
        children: list[Node] = self.tree.root_node.children
        path: list[Node] = [self.tree.root_node]

        while children:
            next_children = []

            for child in children:
                start_line, _ = child.start_point
                end_line, _ = child.end_point
                if start_line <= min_line and max_line <= end_line:
                    path.append(child)
                    next_children.extend(child.children)
                    break

            children = next_children

        start_line, _ = path[-1].start_point
        end_line, _ = path[-1].end_point

        return path

    def get_lines_surrounding(
        self, min_line: int, max_line: int = -1, threshold: int = 4
    ) -> tuple[int, int]:
        if max_line == -1:
            max_line = min_line

        max_num_lines = self.code.count("\n") + 1
        path = self.get_path_to_line(min_line, max_line)
        valid_span = None
        for node in path[::-1]:
            start_line, _ = node.start_point
            end_line, _ = node.end_point
            if (
                0
                < end_line - start_line
                < min(max_line - min_line + threshold, max_num_lines)
            ):
                valid_span = (start_line, end_line)
        if valid_span is not None:
            return valid_span
        else:
            return (min_line, max_line)

    def get_preview(self, min_line: int = 5, max_line: int = -1):
        last_end_line = -1
        lines = self.code.splitlines()
        if max_line == -1:
            # Equation seems to work
            max_line = max(len(lines) // 2 - 200, 50)

        def get_children(node: Node = self.tree.root_node):
            nonlocal last_end_line
            children = []
            for child in node.children:
                start_line, _ = child.start_point
                end_line, _ = child.end_point
                if start_line <= last_end_line:
                    continue
                text = "\n".join(lines[start_line : end_line + 1])
                indentation = " " * (len(text) - len(text.lstrip()))
                for i in range(last_end_line + 1, start_line):
                    line = lines[i]
                    children.append(f"{i} | {line}")
                    last_end_line = i
                if end_line - start_line > max_line:
                    children.extend(get_children(child))
                elif end_line - start_line < min_line:
                    text = "\n".join(
                        [
                            f"{start_line + i} | {line}"
                            for i, line in enumerate(text.split("\n"))
                        ]
                    )
                    children.append(text)
                else:
                    node_lines = text.split("\n")
                    first_line = node_lines[0]
                    first_line = f"{start_line} | {first_line}"
                    second_line = node_lines[1]
                    second_line = f"{start_line + 1} | {second_line}"
                    hidden_lines_content = "\n".join(
                        lines[start_line + 2 : end_line - 1]
                    )
                    number_of_terms = 5
                    first_n_terms = ", ".join(
                        extract_words(hidden_lines_content)[:number_of_terms]
                    )
                    spacing = " " * (len(str(start_line)) + 2)
                    middle_lines = spacing.join(
                        [
                            spacing + indentation + "     ...\n",
                            indentation
                            + f"     (lines {start_line + 1}-{end_line - 1} contains terms: {first_n_terms}\n",
                            indentation + "     ...\n",
                        ]
                    )
                    second_last_line = node_lines[-2]
                    second_last_line = f"{end_line - 1} | {second_last_line}"
                    last_line = node_lines[-1]
                    last_line = f"{end_line} | {last_line}"
                    children.append(first_line)
                    children.append(second_line)
                    children.append(middle_lines)
                    children.append(second_last_line)
                    children.append(last_line)
                last_end_line = end_line
            return children

        return "\n".join(get_children())


def extract_words(string):
    # extract the most common words from a code snippet
    words = re.findall(r"\w+", string)
    return list(dict.fromkeys(words))


def get_global_function_names_and_spans(node):
    return [
        (n.name, (n.lineno, getattr(n, "end_lineno", None)))
        for n in node.body
        if isinstance(n, ast.FunctionDef)
    ]


if __name__ == "__main__":
    snippet = """\
    @patch('os.path.splitext')
    def test_check_comments_presence_with_unsupported_file_extension(self, mock_splitext):
        mock_splitext.return_value = ('file', '.unsupported')
        self.assertEqual(check_comments_presence('file.unsupported', '# This is a comment'), False)"""

    full_code = """\
import unittest
from unittest.mock import patch
from sweepai.utils.comment_utils import check_comments_presence

def helper():
    x = 1
    y = 2
    z = 3
    return x + y + z

class TestCheckCommentsPresence(unittest.TestCase):

    @patch('os.path.splitext')
    def test_check_comments_presence_with_comment(self, mock_splitext):
        mock_splitext.return_value = ('file', '.py')
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        x = 1
        self.assertEqual(check_comments_presence('file.py', '# This is a comment'), True)

    @patch('os.path.splitext')
    def test_check_comments_presence_without_comment(self, mock_splitext):
        mock_splitext.return_value = ('file', '.py')
        self.assertEqual(check_comments_presence('file.py', 'This is not a comment'), False)

    @patch('os.path.splitext')
    def test_check_comments_presence_with_unsupported_file_extension(self, mock_splitext):
        mock_splitext.return_value = ('file', '.unsupported')
        self.assertEqual(check_comments_presence('file.unsupported', '# This is a comment'), False)

    @patch('os.path.splitext')
    def test_check_comments_presence_with_empty_new_code(self, mock_splitext):
        mock_splitext.return_value = ('file', '.py')
        self.assertEqual(check_comments_presence('file.py', ''), False)

if __name__ == '__main__':
    unittest.main()
"""
    # split_code = full_code.split("\n")
    file_contents = open("sweepai/utils/ticket_utils.py").read()
    # file_contents = open("sweepai/handlers/on_ticket.py").read()
    # file_contents = full_code
    # match_start = 16
    # match_end = 20
    code_tree = CodeTree.from_code(file_contents)
    print(code_tree.get_preview())
    print(len(code_tree.get_preview().split("\n")))
    # print(code_tree.get_lines_surrounding(match_start)[0])
    # print(code_tree.get_lines_surrounding(match_end)[1])



from pydantic import BaseModel

from sweepai.core.prompts import (
    diff_section_prompt,
    final_review_prompt,
    human_message_prompt,
    human_message_prompt_comment,
)


class HumanMessagePrompt(BaseModel):
    repo_name: str
    title: str
    summary: str
    snippets: list
    tree: str
    repo_description: str = ""
    snippet_text: str = ""
    commit_history: list = []

    def delete_file(self, file_path):
        # Remove the snippets from the main list
        self.snippets = [
            snippet for snippet in self.snippets if snippet.file_path != file_path
        ]

    def get_relevant_directories(self, directory_tag=None):
        deduped_paths = []
        for snippet in self.snippets:
            if snippet.file_path not in deduped_paths:
                deduped_paths.append(snippet.file_path)
        if len(deduped_paths) == 0:
            return ""
        start_directory_tag = (
            "<relevant_paths_in_repo>" if not directory_tag else f"<{directory_tag}>"
        )
        end_directory_tag = (
            "</relevant_paths_in_repo>" if not directory_tag else f"</{directory_tag}>"
        )
        return (
            start_directory_tag
            + "\n"
            + "\n".join(deduped_paths)
            + "\n"
            + end_directory_tag
        )

    def get_commit_history(self, commit_tag=None):
        return ""
        if len(self.commit_history) == 0:
            return ""
        start_commit_tag = (
            "<relevant_commit_history>" if not commit_tag else f"<{commit_tag}>"
        )
        end_commit_tag = (
            "</relevant_commit_history>" if not commit_tag else f"</{commit_tag}>"
        )
        return (
            start_commit_tag
            + "\n"
            + "\n".join(self.commit_history)
            + "\n"
            + end_commit_tag
        )

    def get_file_paths(self):
        return [snippet.file_path for snippet in self.snippets]

    @staticmethod
    def render_snippet_array(snippets, snippet_tag=None):
        joined_snippets = "\n".join(
            [snippet.get_xml(add_lines=False) for snippet in snippets]
        )
        start_snippet_tag = (
            "<relevant_snippets_in_repo>" if not snippet_tag else f"<{snippet_tag}>"
        )
        end_snippet_tag = (
            "</relevant_snippets_in_repo>" if not snippet_tag else f"</{snippet_tag}>"
        )
        if joined_snippets.strip() == "":
            return ""
        return start_snippet_tag + "\n" + joined_snippets + "\n" + end_snippet_tag

    def render_snippets(self):
        return self.render_snippet_array(self.snippets)

    def construct_prompt(self, snippet_tag=None, directory_tag=None, commit_tag=None):
        relevant_snippets = (
            self.snippet_text
            if self.snippet_text
            else self.render_snippet_array(self.snippets, snippet_tag)
        )
        relevant_directories = (
            self.get_relevant_directories(directory_tag)
            if self.get_relevant_directories(directory_tag)
            else ""
        )
        relevant_commit_history = self.get_commit_history(commit_tag)
        human_messages = [
            {
                "role": msg["role"],
                "content": msg["content"].format(
                    repo_name=self.repo_name,
                    repo_description=self.repo_description,
                    tree=self.tree.strip("\n"),
                    title=self.title,
                    description=(
                        f"Issue Description: {self.summary}"
                        if self.summary.strip()
                        else ""
                    ),
                    relevant_snippets=relevant_snippets,
                    relevant_directories=relevant_directories,
                    relevant_commit_history=relevant_commit_history,
                ),
                "key": msg.get("key"),
            }
            for msg in human_message_prompt
        ]
        return human_messages

    def get_issue_metadata(self):
        self.summary = (
            self.summary if not self.summary.strip().endswith("_No response_") else ""
        )
        issue_description = (
            f"\nIssue Description: {self.summary}" if self.summary else ""
        )
        return f"""# Repo & Issue Metadata
Repo: {self.repo_name}: {self.repo_description}
Issue Title: {self.title}
{issue_description}"""
    
    def get_issue_request(self):
        self.summary = (
            self.summary if not self.summary.strip().endswith("_No response_") else ""
        )
        issue_description = (
            f"\nIssue Description: {self.summary}" if self.summary else ""
        )
        return f"""Issue Title: {self.title}
{issue_description}"""


def render_snippets(snippets):
    res = ""
    for snippet in snippets:
        snippet_text = (
            f"<snippet source={snippet.file_path}>\n{snippet.content}\n</snippet>\n"
        )
        res += snippet_text
    return res

class HumanMessageCommentPrompt(HumanMessagePrompt):
    comment: str
    diffs: list
    relevant_docs: str | None
    pr_file_path: str | None
    pr_chunk: str | None
    original_line: str | None

    def format_diffs(self):
        formatted_diffs = []
        for file_name, file_patch in self.diffs:
            format_diff = diff_section_prompt.format(
                diff_file_path=file_name, diffs=file_patch
            )
            formatted_diffs.append(format_diff)
        return "\n".join(formatted_diffs)

    def construct_prompt(self):
        human_messages = [
            {
                "role": msg["role"],
                "content": msg["content"].format(
                    comment=(
                        self.comment[len("sweep:") :].strip()
                        if self.comment.startswith("sweep:")
                        else self.comment
                    ),
                    repo_name=self.repo_name,
                    repo_description=(
                        self.repo_description if self.repo_description else ""
                    ),
                    diff=self.format_diffs(),
                    title=self.title,
                    tree=self.tree,
                    description=self.summary if self.summary.strip() else "",
                    relevant_directories=self.get_relevant_directories(),
                    relevant_snippets=self.render_snippets(),
                    relevant_commit_history=self.get_commit_history(),
                    relevant_docs=(
                        f"\n{self.relevant_docs}" if self.relevant_docs else ""
                    ),  # conditionally add newline
                ),
            }
            for msg in human_message_prompt_comment
        ]
        return human_messages

    def get_issue_metadata(self):
        self.summary = (
            self.summary if not self.summary.strip().endswith("_No response_") else ""
        )
        issue_description = (
            f"\nIssue Description: {self.summary}" if self.summary.strip() else ""
        )
        return f"""# Repo & Issue Metadata
Repo: {self.repo_name}: {self.repo_description}
Issue Title: {self.title}
{issue_description}
The above was the original plan. Please address the user comment: {self.comment}"""


class HumanMessageFinalPRComment(BaseModel):
    summarization_replies: list

    def construct_prompt(self):
        final_review = final_review_prompt.format(
            file_summaries="\n".join(self.summarization_replies)
        )
        return final_review



import logging
import multiprocessing

import os

from loguru import logger
from tqdm import tqdm

from sweepai.config.client import SweepConfig
from sweepai.core.entities import Snippet
from sweepai.utils.file_utils import read_file_with_fallback_encodings
from sweepai.utils.utils import Tiktoken, chunk_code
from sweepai.utils.timer import Timer

tiktoken_client = Tiktoken()


def filter_file(directory: str, file: str, sweep_config: SweepConfig) -> bool:
    """
    Check if a file should be filtered based on its size and other criteria.

    Args:
        file (str): The path to the file.
        sweep_config (SweepConfig): The configuration object.

    Returns:
        bool: True if the file should be included, False otherwise.
    """
    for ext in sweep_config.exclude_exts:
        if file.endswith(ext):
            return False
    for dir_name in sweep_config.exclude_dirs:
        if file[len(directory) + 1 :].startswith(dir_name):
            return False
    for dir_name in sweep_config.exclude_path_dirs:
        file_parts = file.split(os.path.sep)
        if dir_name in file_parts:
            return False
    try:
        if os.stat(file).st_size > 240000:
            return False
        if os.stat(file).st_size < 10:
            return False
    except FileNotFoundError as e:
        logging.error(f"File not found: {file}. Error: {e}")
        return False
    if not os.path.isfile(file):
        return False
    with open(file, "rb") as f:
        is_binary = False
        for block in iter(lambda: f.read(1024), b""):
            if b"\0" in block:
                is_binary = True
                break
        if is_binary:
            return False
        f.close()
    try:
        # fetch file
        data = read_file_with_fallback_encodings(file)
        lines = data.split("\n")
    except UnicodeDecodeError:
        logger.warning(f"UnicodeDecodeError: {file}, skipping")
        return False
    line_count = len(lines)
    # if average line length is greater than 200, then it is likely not human readable
    if len(data)/line_count > 200:
        return False
    # check token density, if it is greater than 2, then it is likely not human readable
    token_count = tiktoken_client.count(data)
    if token_count == 0:
        return False
    if len(data)/token_count < 2:
        return False
    return True


def read_file(file_name: str) -> str:
    try:
        with open(file_name, "r") as f:
            return f.read()
    except SystemExit:
        raise SystemExit
    except Exception:
        return ""


FILE_THRESHOLD = 240


def file_path_to_chunks(file_path: str) -> list[str]:
    file_contents = read_file(file_path)
    chunks = chunk_code(file_contents, path=file_path)
    return chunks


# @file_cache()
def directory_to_chunks(
    directory: str, sweep_config: SweepConfig
) -> tuple[list[Snippet], list[str]]:
    dir_file_count = {}

    def is_dir_too_big(file_name):
        dir_name = os.path.dirname(file_name)
        only_file_name = os.path.basename(dir_name)
        if only_file_name in ("node_modules", ".venv", "build", "venv", "patch"):
            return True
        if dir_name not in dir_file_count:
            dir_file_count[dir_name] = len(os.listdir(dir_name))
        return dir_file_count[dir_name] > FILE_THRESHOLD

    logger.info(f"Reading files from {directory}")
    vis = set()
    def dfs(file_path: str = directory):
        only_file_name = os.path.basename(file_path)
        if only_file_name in ("node_modules", ".venv", "build", "venv", "patch"):
            return
        if file_path in vis:
            return
        vis.add(file_path)
        if os.path.isdir(file_path):
            for file_name in os.listdir(file_path):
                for sub_file_path in dfs(os.path.join(file_path, file_name)):
                    yield sub_file_path
        else:
            yield file_path
    with Timer():
        file_list = dfs()
        file_list = [
            file_name
            for file_name in file_list
            if filter_file(directory, file_name, sweep_config)
            and os.path.isfile(file_name)
            and not is_dir_too_big(file_name)
        ]
    logger.info("Done reading files")
    all_chunks = []
    with multiprocessing.Pool(processes=multiprocessing.cpu_count() // 4) as pool:
        for chunks in tqdm(pool.imap(file_path_to_chunks, file_list), total=len(file_list)):
            all_chunks.extend(chunks)
    return all_chunks, file_list

if __name__ == "__main__":
    try:
        from sweepai.utils.github_utils import ClonedRepo, get_installation_id
        organization_name = "sweepai"
        
        installation_id = get_installation_id(organization_name)
        cloned_repo = ClonedRepo("sweepai/sweep", installation_id, "main")
        sweep_config = SweepConfig()
        chunks, file_list = directory_to_chunks(cloned_repo.repo_dir, sweep_config)
        # ensure no unallowed files are let through
        assert(not any([file for file in file_list if sweep_config.is_file_excluded(file)]))
        # pick 10 random files and turn them to chunks
        import random
        for _ in range(10):
            idx = random.randint(0, len(file_list) - 1)
            file_chunks = file_path_to_chunks(file_list[idx])

    except Exception as e:
        logger.error(f"repo_parsing_utils.py failed to run successfully with error: {e}")



# Note to contributors: This is an implementation of fuzzy diff.
# It's a bit broken so if you have a better implementation, please replace this.
# Based on https://blog.jcoglan.com/2017/09/19/the-patience-diff-algorithm/

# TODO: Improve this with Hungarian method to find the best matches
# treating the matches as an assignment problem.
# Minimize the product of the fuzziness scores. We can take the negative
# log of the fuzziness scores and use that as the cost matrix.

from rapidfuzz import fuzz

THRESHOLD = 95

def similar(a, b):
    return fuzz.ratio(a, b) > THRESHOLD

def lis(lst: list):
    # Longest increasing subsequence
    # Patience sorting algorithm
    if not lst:
        return []

    dag = {lst[0]: None}
    buckets = [[lst[0]]]
    
    for i, x in enumerate(lst[1:]):
        # Should use binary search here for O(nlogn) but quadratic for now
        # Haven't seen performance issues yet
        for j, bucket in enumerate(buckets):
            if x < bucket[-1]:
                buckets[j].append(x)
                dag[x] = buckets[j - 1][-1] if j > 0 else None
                break
        else:
            buckets.append([x])
            dag[x] = buckets[-2][-1]
        
    result = []
    k = buckets[-1][-1]
    while k is not None:
        result.append(k)
        k = dag[k]

    back_map = {x: i for i, x in enumerate(lst)}
    return [back_map[x] for x in result[::-1]]

def find_unique_matches(
    old_lines: list,
    new_lines: list
):
    matched_lines = set()
    matches = []
    for i, old_line in enumerate(old_lines):
        if old_line in old_lines[:i] or old_line in old_lines[i + 1:]:
            continue
        max_index = -1
        max_fuzz = 0
        for j, new_line in enumerate(new_lines):
            if j in matched_lines:
                continue
            ratio = fuzz.ratio(old_line, new_line)
            if ratio > max_fuzz:
                max_fuzz = ratio
                max_index = j
        if max_fuzz > THRESHOLD:
            matches.append((i, max_index))
            matched_lines.add(max_index)
    return matches


def patience_fuzzy_diff_lines(
    old_lines: list[str],
    new_lines: list[str]
) -> str:
    # Assumes new string has a few lines added from old string.
    # There's probably a better implementation but we'll do this for now.
    if not old_lines:
        return [f"+ {line}" for line in new_lines]
    if not new_lines:
        return [f"- {line}" for line in old_lines]
    if len(old_lines) == len(new_lines) and all(similar(old_line, new_line) for old_line, new_line in zip(old_lines, new_lines)):
        return [f"  {line}" for line in new_lines]
    diff_lines = []
    matches = find_unique_matches(old_lines, new_lines)
    match_indices = lis([right for _left, right in matches])
    matches = [matches[i] for i in match_indices]
    if matches:
        last_left, last_right = 0, 0
        for left, right in matches:
            current_left_lines = old_lines[last_left:left]
            current_right_lines = new_lines[last_right:right]

            diff_lines.extend(patience_fuzzy_diff_lines(current_left_lines, current_right_lines))
            diff_lines.append(f"  {new_lines[right]}")

            last_left = left + 1
            last_right = right + 1
        diff_lines.extend(patience_fuzzy_diff_lines(old_lines[last_left:], new_lines[last_right:]))
    else:
        if similar(old_lines[0], new_lines[0]):
            diff_lines.append(f"  {new_lines[0]}")
        else:
            diff_lines.append(f"- {old_lines[0]}")
            diff_lines.append(f"+ {new_lines[0]}")
        diff_lines.extend(patience_fuzzy_diff_lines(old_lines[1:], new_lines[1:]))
    return diff_lines

def patience_fuzzy_diff(
    old_string: str,
    new_string: str
):
    if old_string == new_string:
        return ""
    old_lines = old_string.splitlines()
    new_lines = new_string.splitlines()
    diff_lines = patience_fuzzy_diff_lines(old_lines, new_lines)
    return "\n".join(diff_lines)

def patience_fuzzy_additions(
    old_string: str,
    new_string: str
):
    if old_string == new_string:
        return ""
    old_lines = old_string.splitlines()
    new_lines = new_string.splitlines()
    diff_lines = patience_fuzzy_diff_lines(old_lines, new_lines)
    return "\n".join(line[2:] for line in diff_lines if line.startswith("+"))

old_lint_results = """> pylint sweepai/handlers/on_ticket.py

************* Module on_ticket
sweepai/handlers/on_ticket.py:167:52: W0613: Unused argument 'reaction_content' (unused-argument)
sweepai/handlers/on_ticket.py:221:24: W3101: Missing timeout argument for method 'requests.get' can cause your program to hang indefinitely (missing-timeout)
sweepai/handlers/on_ticket.py:251:24: W3101: Missing timeout argument for method 'requests.get' can cause your program to hang indefinitely (missing-timeout)
sweepai/handlers/on_ticket.py:268:34: W0612: Unused variable 'user_message' (unused-variable)
sweepai/handlers/on_ticket.py:525:16: W0719: Raising too general exception: Exception (broad-exception-raised)"""

new_lint_results = """> pylint sweepai/handlers/on_ticket.py

************* Module on_ticket
sweepai/handlers/on_ticket.py:167:52: W0613: Unused argument 'reaction_content' (unused-argument)
sweepai/handlers/on_ticket.py:245:12: W0621: Redefining name 'logger' from outer scope (line 23) (redefined-outer-name)
sweepai/handlers/on_ticket.py:221:24: W3101: Missing timeout argument for method 'requests.get' can cause your program to hang indefinitely (missing-timeout)
sweepai/handlers/on_ticket.py:245:21: E0602: Undefined variable 'Logger' (undefined-variable)
sweepai/handlers/on_ticket.py:252:24: W3101: Missing timeout argument for method 'requests.get' can cause your program to hang indefinitely (missing-timeout)
sweepai/handlers/on_ticket.py:269:34: W0612: Unused variable 'user_message' (unused-variable)
sweepai/handlers/on_ticket.py:526:16: W0719: Raising too general exception: Exception (broad-exception-raised)"""

# old_lint_results = """David Axelrod
# Electric Prunes
# Gil Scott Heron
# The Slits
# Faust
# The Sonics
# The Sonics"""

# new_lint_results = """The Slits
# Electric Prunes
# Gil Scott Heron
# David Axelrod
# Electric Prunes
# Faust
# The Sonics
# The Sonics"""

stress_test_old = "AAGTCCGTAACCTGACATCTGAGGCTAATCACTGAGGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGAT"
stress_test_new = "AAGTCCGTAACCTGACATCTGAGGCTAATCACTGAGGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGATATGCGTATGCGCGAT"

if __name__ == "__main__":
    import time
    start = time.time()
    print(patience_fuzzy_diff(
        old_lint_results,
        new_lint_results
    ))
    # print(patience_fuzzy_diff_lines(
    #     stress_test_old,
    #     stress_test_new
    # ))
    print(f"Time taken: {time.time() - start} seconds.")



from sweepai.config.client import SweepConfig

# post process rip grep output to be more condensed
def post_process_rg_output(root_directory: str, sweep_config: SweepConfig, output: str):
    processed_output = ""
    output_lines = output.split("\n")
    # empty lines are present at end of output
    output_lines = [line for line in output_lines if line]
    file_output_dict = {}
    file_to_num_occurrences = {}
    for line in output_lines:
        filename, content = line.split(":", 1)
        filename = filename[len(root_directory) + 1:]
        if not sweep_config.is_file_excluded_aggressive(root_directory, filename):
            if filename not in file_output_dict:
                file_output_dict[filename] = ""
            file_output_dict[filename] += (content + "\n")
            if filename not in file_to_num_occurrences:
                file_to_num_occurrences[filename] = 0
            file_to_num_occurrences[filename] += 1
    
    # determine if we need to truncate the output
    total_output_length = sum([len(line) for content in file_output_dict.values() for line in content])
    file_name_and_contents = [(filename, content) for filename, content in file_output_dict.items()]
    file_name_and_contents.sort(key=lambda x: x[0])
    if total_output_length > sweep_config.truncation_cutoff:
        for filename, content in file_name_and_contents:
            processed_output += f"File: {filename} contained the following matching lines of code"
            content = content.split("\n")
            if len(content) < 4:
                processed_output += " :\n"
                for line in content:
                    processed_output += f"{line}\n"
            else:
                processed_output += " (truncated):\n"
                line1 = content[0]
                line2 = content[-1]
                if len(line1) > 200:
                    line1 = line1[:20] + " ..."
                if len(line2) > 200:
                    line2 = line2[:20] + " ..."
                processed_output += f"{line1}\n"
                processed_output += "...\n"
                processed_output += f"{line2}\n"
            processed_output += "\n"
    else:
        for filename, content in file_name_and_contents:
            processed_output += f"File: {filename} contained the following matching lines of code:\n" + content + "\n"
    return processed_output, file_output_dict, file_to_num_occurrences

# try and find code_snippet inside file_contents given various levels of indentation, and right strip the lines of code
# if successful returns the num of spaces required to find the code match and if we need to rstrip the old code or not
def manual_code_check(file_contents: str, code_snippet: str) -> tuple[int, bool]:
    code_lines = [line for line in code_snippet.split("\n")]
    # special case for single line
    if len(code_lines) == 1:
        file_lines = file_contents.split("\n")
        # check directly
        new_code = code_lines[0]
        # only continue if it is unique, this will then later fail the uniqueness check
        if file_contents.count(new_code) > 1:
            return 0, False
        if new_code in file_contents:
            # now check how many leading whitespaces there are
            for line in file_lines:
                if new_code in line:
                    return len(line)-len(line.lstrip()), False
        else:
            # now try rstrip if initially the code is not there
            new_code = new_code.rstrip()
            if file_contents.count(new_code) > 1: # uniqueness check
                return 0, False
            if new_code in file_contents:
               # now check how many leading whitespaces there are
                for line in file_lines:
                    if new_code in line:
                        return len(line)-len(line.lstrip()), False
        return -1, False
                
    # assume one indent is two spaces and check max 10 indents
    for indent in range(0, 40, 2):
        new_code_lines = [f"{' ' * indent}{line}" for line in code_lines]
        new_code = "\n".join(new_code_lines)
        if new_code in file_contents:
            return indent, False
    # sometimes llm returns code with trailing whitespace, if we have reached here check again but strip all trailing whitespace
    code_lines = [line.rstrip() for line in code_snippet.split("\n")]
    for indent in range(0, 40, 2):
        new_code_lines = [f"{' ' * indent}{line}" for line in code_lines]
        new_code = "\n".join(new_code_lines)
        if new_code in file_contents:
            return indent, True
    return -1, False


import hashlib
import os
import re
import time

from sweepai.config.server import PROGRESS_BASE_URL

UPDATES_MESSAGE = f"""\
<details>
<summary><b>🎉 Latest improvements to Sweep:</b></summary>
<ul>
<li>New <a href="{PROGRESS_BASE_URL}">dashboard</a> launched for real-time tracking of Sweep issues, covering all stages from search to coding.</li>
<li>Integration of OpenAI's latest Assistant API for more efficient and reliable code planning and editing, improving speed by 3x.</li>
<li>Use the <a href="https://marketplace.visualstudio.com/items?itemName=GitHub.vscode-pull-request-github">GitHub issues extension</a> for creating Sweep issues directly from your editor.</li>
</ul>
</details>
"""

DEFAULT_BOT_SUFFIX = "\n\n*This is an automated message generated by [Sweep AI](https://sweep.dev).*"
BOT_SUFFIX = os.environ.get("BOT_SUFFIX", DEFAULT_BOT_SUFFIX).replace("\\n", "\n")

if BOT_SUFFIX != DEFAULT_BOT_SUFFIX:
    print(f"Using custom bot suffix: {BOT_SUFFIX}")

FASTER_MODEL_MESSAGE = """\
You ran out of the free tier GPT-4 tickets! We no longer support running Sweep with GPT-3.5 as it is too unreliable. Here are your options:
- You can get a free trial of Sweep Pro to get unlimited GPT-4 tickets [here](https://buy.stripe.com/00g5npeT71H2gzCfZ8) or purchase a individual GPT-4 tickets [here](https://buy.stripe.com/00g3fh7qF85q0AE14d).
- You can self-host Sweep with your own GPT-4 API key. You can find instructions [here](https://docs.sweep.dev/deployment).
- You can book a chat with us to discuss your use case and get additional free GPT-4 tickets [here](https://calendly.com/d/2n5-3qf-9xy/user-interview).
"""

sep = "\n---\n"
bot_suffix_starring = ""
bot_suffix = (
    f"\n{sep}\n{UPDATES_MESSAGE}\n\n💡 To recreate the pull request edit the issue"
    " title or description."
)
discord_suffix = (
    "\n<sup>Something wrong? [Let us know](https://discord.gg/sweep).</sup>"
)

stars_suffix = ""

collapsible_template = """
<details {opened}>
<summary>{summary}</summary>

{body}
</details>
"""

checkbox_template = "- [{check}] {filename}\n{instructions}\n"

num_of_snippets_to_query = 30
total_number_of_snippet_tokens = 15_000
num_full_files = 2


def ordinal(n):
    return str(n) + (
        "th" if 4 <= n <= 20 else {1: "st", 2: "nd", 3: "rd"}.get(n % 10, "th")
    )


def format_sandbox_success(success):
    return "✓" if success else "❌ (`Sandbox Failed`)"


def create_collapsible(summary: str, body: str, opened: bool = False):
    return collapsible_template.format(
        summary=summary, body=body, opened="open" if opened else ""
    )


def inline_code(text: str):
    return f"<code>{text}</code>" if text else ""


def code_block(text: str):
    return f"<pre>{text}</pre>" if text else ""


def blockquote(text: str):
    text = text.replace("\n•", "<br/>•")
    return f"<blockquote>{text}\n</blockquote>" if text else ""


def create_checkbox(title: str, body: str, checked: bool = False):
    return checkbox_template.format(
        check="X" if checked else " ", filename=title, instructions=body
    )


def strip_sweep(text: str):
    return (
        re.sub(
            r"^[Ss]weep\s?(\([Ss]low\))?(\([Mm]ap\))?(\([Ff]ast\))?\s?:", "", text
        ).lstrip(),
        re.search(r"^[Ss]weep\s?\([Ss]low\)", text) is not None,
        re.search(r"^[Ss]weep\s?\([Mm]ap\)", text) is not None,
        re.search(r"^[Ss]weep\s?\([Ss]ubissues?\)", text) is not None,
        re.search(r"^[Ss]weep\s?\([Ss]andbox?\)", text) is not None,
        re.search(r"^[Ss]weep\s?\([Ff]ast\)", text) is not None,
        re.search(r"^[Ss]weep\s?\([Ll]int\)", text) is not None,
    )


def clean_logs(logs: str):
    cleaned_logs = re.sub(r"\x1b\[.*?[@-~]", "", logs.replace("```", "\`\`\`"))
    cleaned_logs = re.sub("\n{2,}", "\n", cleaned_logs)
    cleaned_logs = re.sub("\r{2,}", "\n", cleaned_logs)
    cleaned_logs = cleaned_logs.strip("\n")
    cleaned_logs = cleaned_logs or "(nothing was outputted)"
    return cleaned_logs


def extract_lines(text: str, start: int, end: int):
    lines = text.splitlines(keepends=True)
    return "\n".join(lines[max(0, start) : min(len(lines), end)])


def add_line_numbers(text: str, start: int = 0):
    lines = text.splitlines(keepends=True)
    return "".join(f"{start + i} | {line}" for i, line in enumerate(lines))


def to_branch_name(s, max_length=40):
    branch_name = s.strip().lower().replace(" ", "_")
    branch_name = re.sub(r"[^a-z0-9_]", "", branch_name)
    return branch_name[:max_length]


def get_hash():
    return hashlib.sha256(str(time.time()).encode()).hexdigest()[:10]

# used for getting all indices of a substring match
def get_all_indices_of_substring(content: str, substring: str):
    start = 0
    indices = []
    while True:
        index = content.find(substring, start)
        if index == -1:  # No more occurrences found
            break
        indices.append(index)
        start = index + 1  # Move past the last found occurrence
    return indices


from datetime import datetime


def compute_score(relative_file_path, git_repo):
    commits = list(git_repo.iter_commits(paths=relative_file_path))
    score_factor = get_factors(commits)
    return score_factor


def get_factors(commits):
    commit_count = len(commits) + 1
    earliest_commit = commits[0].committed_datetime if commits else datetime.now()
    current_time = datetime.now()
    tz_info = earliest_commit.astimezone().tzinfo
    if tz_info:
        current_time = datetime.now().astimezone(tz_info)
        earliest_commit = earliest_commit.astimezone(tz_info)
    days_since_last_modified = (
        max(
            ((current_time - earliest_commit).total_seconds() // 3600),
            0,
        )
        + 1
    )
    return (1, commit_count, days_since_last_modified)


def convert_to_percentiles(values, max_percentile=0.25):
    sorted_values = sorted(values)  # Sort the values in ascending order
    n = len(sorted_values)
    percentile_mapping = {
        value: (i / (n)) * max_percentile for i, value in enumerate(sorted_values)
    }
    percentiles = [
        percentile_mapping[value] for value in values
    ]  # Create the percentiles list based on the mapping

    return percentiles


def get_scores(score_factors):
    # add all of the arrays together
    line_count_scores = [x[0] for x in score_factors]
    commit_count_scores = [x[1] for x in score_factors]
    days_since_last_modified_scores = [x[2] for x in score_factors]
    line_count_scores = convert_to_percentiles(line_count_scores, 1)
    commit_count_scores = convert_to_percentiles(commit_count_scores, 1)
    days_since_last_modified_scores = convert_to_percentiles(
        days_since_last_modified_scores, 1
    )
    days_since_last_modified_scores = [1 - score for score in days_since_last_modified_scores]
    scores = [
        sum(x)
        for x in zip(
            line_count_scores, commit_count_scores, days_since_last_modified_scores
        )
    ]
    return convert_to_percentiles(scores, 0.25)


import linecache
import sys


def trace_lines(frame, event, arg):
    if event == "line":
        filename = frame.f_code.co_filename
        if "" in filename:
            lineno = frame.f_lineno
            line = linecache.getline(filename, lineno)
            print(f"Executing {filename}:line {lineno}:{line.rstrip()}")
    return trace_lines


sys.settrace(trace_lines)



import json
import unittest

from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)

from sweepai.agents.assistant_wrapper import fix_tool_calls

if __name__ == "__main__":
    tool_call = ChatCompletionMessageToolCall(
        id="tool_call_id",
        function=Function(
            arguments="arguments",
            name="function_name",
        ),
        type="function",
    )


class TestFixToolCalls(unittest.TestCase):
    def test_multiple_tool_calls(self):
        # Setup input with multiple tool calls, including more than one 'parallel' tool calls
        input_tool_calls = [
            ChatCompletionMessageToolCall(
                id="1",
                type="function",
                function={
                    "name": "parallel",
                    "arguments": json.dumps(
                        {
                            "tool_uses": [
                                {
                                    "recipient_name": "functions.example_function",
                                    "parameters": {"arg1": "value1"},
                                },
                                {
                                    "recipient_name": "functions.example_function",
                                    "parameters": {"arg1": "value2"},
                                },
                            ]
                        }
                    ),
                },
            ),
            ChatCompletionMessageToolCall(
                id="2",
                type="function",
                function={
                    "name": "example_tool",
                    "arguments": json.dumps({"arg2": "value2"}),
                },
            ),
            ChatCompletionMessageToolCall(
                id="3",
                type="function",
                function={
                    "name": "parallel",
                    "arguments": json.dumps(
                        {
                            "tool_uses": [
                                {
                                    "recipient_name": "functions.another_function",
                                    "parameters": {"arg3": "value3"},
                                }
                            ]
                        }
                    ),
                },
            ),
        ]

        # Expected tool calls after fix
        expected_tool_calls = [
            ChatCompletionMessageToolCall(
                id="1_0",
                type="function",
                function={
                    "name": "example_function",
                    "arguments": json.dumps({"arg1": "value1"}),
                },
            ),
            ChatCompletionMessageToolCall(
                id="1_1",
                type="function",
                function={
                    "name": "example_function",
                    "arguments": json.dumps({"arg1": "value2"}),
                },
            ),
            ChatCompletionMessageToolCall(
                id="2",
                type="function",
                function={
                    "name": "example_tool",
                    "arguments": json.dumps({"arg2": "value2"}),
                },
            ),
            ChatCompletionMessageToolCall(
                id="3_0",
                type="function",
                function={
                    "name": "another_function",
                    "arguments": json.dumps({"arg3": "value3"}),
                },
            ),
        ]

        # Run the fix_tool_calls function
        output_tool_calls = fix_tool_calls(input_tool_calls)
        self.assertEqual(len(output_tool_calls), len(expected_tool_calls))
        for actual, expected in zip(output_tool_calls, expected_tool_calls):
            self.assertEqual(actual.id, expected.id)
            self.assertEqual(actual.type, expected.type)
            self.assertEqual(actual.function, expected.function)


if __name__ == "__main__":
    unittest.main()



import base64
import os

from dotenv import load_dotenv
from loguru import logger

logger.print = logger.info

load_dotenv(dotenv_path=".env", override=True, verbose=True)

os.environ["GITHUB_APP_PEM"] = os.environ.get("GITHUB_APP_PEM") or base64.b64decode(
    os.environ.get("GITHUB_APP_PEM_BASE64", "")
).decode("utf-8")

if os.environ["GITHUB_APP_PEM"]:
    os.environ["GITHUB_APP_ID"] = (
        (os.environ.get("GITHUB_APP_ID") or os.environ.get("APP_ID"))
        .replace("\\n", "\n")
        .strip('"')
    )

os.environ["TRANSFORMERS_CACHE"] = os.environ.get(
    "TRANSFORMERS_CACHE", "/tmp/cache/model"
)  # vector_db.py
os.environ["TIKTOKEN_CACHE_DIR"] = os.environ.get(
    "TIKTOKEN_CACHE_DIR", "/tmp/cache/tiktoken"
)  # utils.py

SENTENCE_TRANSFORMERS_MODEL = os.environ.get(
    "SENTENCE_TRANSFORMERS_MODEL",
    "sentence-transformers/all-MiniLM-L6-v2",  # "all-mpnet-base-v2"
)

TEST_BOT_NAME = "sweep-nightly[bot]"
ENV = os.environ.get("ENV", "dev")
# ENV = os.environ.get("MODAL_ENVIRONMENT", "dev")

# ENV = PREFIX
# ENVIRONMENT = PREFIX

DB_MODAL_INST_NAME = "db"
DOCS_MODAL_INST_NAME = "docs"
API_MODAL_INST_NAME = "api"
UTILS_MODAL_INST_NAME = "utils"

BOT_TOKEN_NAME = "bot-token"

# goes under Modal 'discord' secret name (optional, can leave env var blank)
DISCORD_WEBHOOK_URL = os.environ.get("DISCORD_WEBHOOK_URL")
DISCORD_MEDIUM_PRIORITY_URL = os.environ.get("DISCORD_MEDIUM_PRIORITY_URL")
DISCORD_LOW_PRIORITY_URL = os.environ.get("DISCORD_LOW_PRIORITY_URL")
DISCORD_FEEDBACK_WEBHOOK_URL = os.environ.get("DISCORD_FEEDBACK_WEBHOOK_URL")

SWEEP_HEALTH_URL = os.environ.get("SWEEP_HEALTH_URL")
DISCORD_STATUS_WEBHOOK_URL = os.environ.get("DISCORD_STATUS_WEBHOOK_URL")

# goes under Modal 'github' secret name
GITHUB_APP_ID = os.environ.get("GITHUB_APP_ID", os.environ.get("APP_ID"))
# deprecated: old logic transfer so upstream can use this
if GITHUB_APP_ID is None:
    if ENV == "prod":
        GITHUB_APP_ID = "307814"
    elif ENV == "dev":
        GITHUB_APP_ID = "324098"
    elif ENV == "staging":
        GITHUB_APP_ID = "327588"
GITHUB_BOT_USERNAME = os.environ.get("GITHUB_BOT_USERNAME")

# deprecated: left to support old logic
if not GITHUB_BOT_USERNAME:
    if ENV == "prod":
        GITHUB_BOT_USERNAME = "sweep-ai[bot]"
    elif ENV == "dev":
        GITHUB_BOT_USERNAME = "sweep-nightly[bot]"
    elif ENV == "staging":
        GITHUB_BOT_USERNAME = "sweep-canary[bot]"
elif not GITHUB_BOT_USERNAME.endswith("[bot]"):
    GITHUB_BOT_USERNAME = GITHUB_BOT_USERNAME + "[bot]"

GITHUB_LABEL_NAME = os.environ.get("GITHUB_LABEL_NAME", "sweep")
GITHUB_LABEL_COLOR = os.environ.get("GITHUB_LABEL_COLOR", "9400D3")
GITHUB_LABEL_DESCRIPTION = os.environ.get(
    "GITHUB_LABEL_DESCRIPTION", "Sweep your software chores"
)
GITHUB_APP_PEM = os.environ.get("GITHUB_APP_PEM")
GITHUB_APP_PEM = GITHUB_APP_PEM or os.environ.get("PRIVATE_KEY")
if GITHUB_APP_PEM is not None:
    GITHUB_APP_PEM = GITHUB_APP_PEM.strip(' \n"')  # Remove whitespace and quotes
    GITHUB_APP_PEM = GITHUB_APP_PEM.replace("\\n", "\n")

GITHUB_CONFIG_BRANCH = os.environ.get("GITHUB_CONFIG_BRANCH", "sweep/add-sweep-config")
GITHUB_DEFAULT_CONFIG = os.environ.get(
    "GITHUB_DEFAULT_CONFIG",
    """# Sweep AI turns bugs & feature requests into code changes (https://sweep.dev)
# For details on our config file, check out our docs at https://docs.sweep.dev/usage/config

# This setting contains a list of rules that Sweep will check for. If any of these rules are broken in a new commit, Sweep will create an pull request to fix the broken rule.
rules:
{additional_rules}

# This is the branch that Sweep will develop from and make pull requests to. Most people use 'main' or 'master' but some users also use 'dev' or 'staging'.
branch: 'main'

# By default Sweep will read the logs and outputs from your existing Github Actions. To disable this, set this to false.
gha_enabled: True

# This is the description of your project. It will be used by sweep when creating PRs. You can tell Sweep what's unique about your project, what frameworks you use, or anything else you want.
#
# Example:
#
# description: sweepai/sweep is a python project. The main api endpoints are in sweepai/api.py. Write code that adheres to PEP8.
description: ''

# This sets whether to create pull requests as drafts. If this is set to True, then all pull requests will be created as drafts and GitHub Actions will not be triggered.
draft: False

# This is a list of directories that Sweep will not be able to edit.
blocked_dirs: []
""",
)


MONGODB_URI = os.environ.get("MONGODB_URI", None)
IS_SELF_HOSTED = os.environ.get("IS_SELF_HOSTED", "true").lower() == "true"

REDIS_URL = os.environ.get("REDIS_URL")
if not REDIS_URL:
    REDIS_URL = os.environ.get("redis_url", "redis://0.0.0.0:6379/0")

ORG_ID = os.environ.get("ORG_ID", None)
POSTHOG_API_KEY = os.environ.get(
    "POSTHOG_API_KEY", "phc_CnzwIB0W548wN4wEGeRuxXqidOlEUH2AcyV2sKTku8n"
)

E2B_API_KEY = os.environ.get("E2B_API_KEY")

SUPPORT_COUNTRY = os.environ.get("GDRP_LIST", "").split(",")

WHITELISTED_REPOS = os.environ.get("WHITELISTED_REPOS", "").split(",")
BLACKLISTED_USERS = os.environ.get("BLACKLISTED_USERS", "").split(",")


os.environ["TOKENIZERS_PARALLELISM"] = "false"

ACTIVELOOP_TOKEN = os.environ.get("ACTIVELOOP_TOKEN", None)

VECTOR_EMBEDDING_SOURCE = os.environ.get(
    "VECTOR_EMBEDDING_SOURCE", "openai"
)  # Alternate option is openai or huggingface and set the corresponding env vars

BASERUN_API_KEY = os.environ.get("BASERUN_API_KEY", None)

# Huggingface settings, only checked if VECTOR_EMBEDDING_SOURCE == "huggingface"
HUGGINGFACE_URL = os.environ.get("HUGGINGFACE_URL", None)
HUGGINGFACE_TOKEN = os.environ.get("HUGGINGFACE_TOKEN", None)

# Replicate settings, only checked if VECTOR_EMBEDDING_SOURCE == "replicate"
REPLICATE_API_KEY = os.environ.get("REPLICATE_API_KEY", None)
REPLICATE_URL = os.environ.get("REPLICATE_URL", None)
REPLICATE_DEPLOYMENT_URL = os.environ.get("REPLICATE_DEPLOYMENT_URL", None)

# Default OpenAI
OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY", None)

OPENAI_API_TYPE = os.environ.get("OPENAI_API_TYPE", "anthropic")
assert OPENAI_API_TYPE in ["anthropic", "azure", "openai"], "Invalid OPENAI_API_TYPE"
OPENAI_EMBEDDINGS_API_TYPE = os.environ.get("OPENAI_EMBEDDINGS_API_TYPE", "openai")

AZURE_API_KEY = os.environ.get("AZURE_API_KEY", None)
OPENAI_API_BASE = os.environ.get("OPENAI_API_BASE", None)
OPENAI_API_VERSION = os.environ.get("OPENAI_API_VERSION", None)
AZURE_OPENAI_DEPLOYMENT = os.environ.get("AZURE_OPENAI_DEPLOYMENT", None)

OPENAI_EMBEDDINGS_API_TYPE = os.environ.get("OPENAI_EMBEDDINGS_API_TYPE", "openai")
OPENAI_EMBEDDINGS_AZURE_ENDPOINT = os.environ.get(
    "OPENAI_EMBEDDINGS_AZURE_ENDPOINT", None
)
OPENAI_EMBEDDINGS_AZURE_API_KEY = os.environ.get(
    "OPENAI_EMBEDDINGS_AZURE_API_KEY", None
)
OPENAI_EMBEDDINGS_AZURE_DEPLOYMENT = os.environ.get(
    "OPENAI_EMBEDDINGS_AZURE_DEPLOYMENT", None
)
OPENAI_EMBEDDINGS_AZURE_API_VERSION = os.environ.get(
    "OPENAI_EMBEDDINGS_AZURE_API_VERSION", None
)

OPENAI_API_ENGINE_GPT35 = os.environ.get("OPENAI_API_ENGINE_GPT35", None)
OPENAI_API_ENGINE_GPT4 = os.environ.get("OPENAI_API_ENGINE_GPT4", None)
OPENAI_API_ENGINE_GPT4_32K = os.environ.get("OPENAI_API_ENGINE_GPT4_32K", None)
MULTI_REGION_CONFIG = os.environ.get("MULTI_REGION_CONFIG", None)
if isinstance(MULTI_REGION_CONFIG, str):
    MULTI_REGION_CONFIG = MULTI_REGION_CONFIG.strip("'").replace("\\n", "\n")
    MULTI_REGION_CONFIG = [item.split(",") for item in MULTI_REGION_CONFIG.split("\n")]

WHITELISTED_USERS = os.environ.get("WHITELISTED_USERS", None)
if WHITELISTED_USERS:
    WHITELISTED_USERS = WHITELISTED_USERS.split(",")
    WHITELISTED_USERS.append(GITHUB_BOT_USERNAME)

DEFAULT_GPT4_32K_MODEL = os.environ.get("DEFAULT_GPT4_32K_MODEL", "gpt-4-0125-preview")
DEFAULT_GPT35_MODEL = os.environ.get("DEFAULT_GPT35_MODEL", "gpt-3.5-turbo-1106")

RESEND_API_KEY = os.environ.get("RESEND_API_KEY", None)
LOKI_URL = None

DEBUG = os.environ.get("DEBUG", "false").lower() == "true"
ENV = "prod" if GITHUB_BOT_USERNAME != TEST_BOT_NAME else "dev"

PROGRESS_BASE_URL = os.environ.get(
    "PROGRESS_BASE_URL", "https://progress.sweep.dev"
).rstrip("/")

DISABLED_REPOS = os.environ.get("DISABLED_REPOS", "").split(",")

GHA_AUTOFIX_ENABLED: bool = os.environ.get("GHA_AUTOFIX_ENABLED", False)
MERGE_CONFLICT_ENABLED: bool = os.environ.get("MERGE_CONFLICT_ENABLED", False)
INSTALLATION_ID = os.environ.get("INSTALLATION_ID", None)

AWS_ACCESS_KEY=os.environ.get("AWS_ACCESS_KEY")
AWS_SECRET_KEY=os.environ.get("AWS_SECRET_KEY")
AWS_REGION=os.environ.get("AWS_REGION")
ANTHROPIC_AVAILABLE = AWS_ACCESS_KEY and AWS_SECRET_KEY and AWS_REGION

USE_ASSISTANT = os.environ.get("USE_ASSISTANT", "true").lower() == "true"
ANTHROPIC_API_KEY = os.environ.get("ANTHROPIC_API_KEY", None)

COHERE_API_KEY = os.environ.get("COHERE_API_KEY", None)

VOYAGE_API_KEY = os.environ.get("VOYAGE_API_KEY", None)

VOYAGE_API_AWS_ACCESS_KEY=os.environ.get("VOYAGE_API_AWS_ACCESS_KEY_ID")
VOYAGE_API_AWS_SECRET_KEY=os.environ.get("VOYAGE_API_AWS_SECRET_KEY")
VOYAGE_API_AWS_REGION=os.environ.get("VOYAGE_API_AWS_REGION")
VOYAGE_API_AWS_ENDPOINT_NAME=os.environ.get("VOYAGE_API_AWS_ENDPOINT_NAME", "voyage-code-2")

VOYAGE_API_USE_AWS = VOYAGE_API_AWS_ACCESS_KEY and VOYAGE_API_AWS_SECRET_KEY and VOYAGE_API_AWS_REGION

PAREA_API_KEY = os.environ.get("PAREA_API_KEY", None)

# TODO: we need to make this dynamic + backoff
BATCH_SIZE = int(
    os.environ.get("BATCH_SIZE", 64 if VOYAGE_API_KEY else 256) # Voyage only allows 128 items per batch and 120000 tokens per batch
)

DEPLOYMENT_GHA_ENABLED = os.environ.get("DEPLOYMENT_GHA_ENABLED", "true").lower() == "true"

JIRA_USER_NAME = os.environ.get("JIRA_USER_NAME", None)
JIRA_API_TOKEN = os.environ.get("JIRA_API_TOKEN", None)
JIRA_URL = os.environ.get("JIRA_URL", None)



import queue
import threading


class SafePriorityQueue:
    def __init__(self):
        self.q = queue.PriorityQueue()
        self.lock = threading.Lock()

    def put(self, priority: int, event):
        try:
            with self.lock:
                self.q.put((priority, event))
                self._invalidate_lower_priority(priority)
        except Exception:
            pass

    def get(self):
        with self.lock:
            return self.q.get()[1]  # Only return the event, not the priority

    def empty(self):
        with self.lock:
            return self.q.empty()

    def _invalidate_lower_priority(self, priority: int):
        temp_q = queue.PriorityQueue()
        while not self.q.empty():
            p, e = self.q.get()
            if p <= priority:
                temp_q.put((p, e))
        self.q = temp_q
